{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Actions.Backup (
    autoBackup
  , backupPath
  , checkBackup
  , restorePath
  , symlinkPath
  ) where

import F4L
import F4L.Actions.Common
import F4L.Actions.Files
import F4L.Actions.Ops
import F4L.Actions.Storages
import F4L.BufferMonad
import F4L.F4LMonad
import F4L.FidStore
import F4L.FileTree as FT
import F4L.FSMonad hiding ( outputLineMsg, cleanLine, symlink )
import F4L.Knowledge
import F4L.Op
import F4L.StoryTree as ST
import F4L.Utils hiding( fileSize )

import Lens.Micro.Platform
import Control.Monad
import Control.Monad.Trans.Except( ExceptT(..), runExceptT )
import Control.Monad.Reader
import Control.Monad.State.Strict
import qualified Data.IntMap.Strict as IM
import Data.List( sortBy )
import qualified Data.Map.Strict as M -- TODO: just use SM
import qualified Data.Map.Strict as SM
import qualified Data.Set as S
import System.FilePath( combine, takeDirectory, joinPath, splitDirectories )


-- | Backup at most maxSize bytes depending on priority, size and known copies
autoBackup :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k ) 
              => FilePath -> Int -> F4LMonad k m ()
autoBackup dest maxSize = do
  fullData <- use st_knowledge
  fidstore <- use $ st_opEnv . openv_fids
  -- Fids to backup by reverse weight
  let fids = map fst $ sortBy ( compare `byRev` benefitOverCost ) $ map ( mapFst $ fidByIndex' fidstore ) $ IM.toList $ view kd_files fullData
  backup fids dest maxSize

  where
    -- TODO: find a more reasonable fun
    benefitOverCost :: ( FileId, FileData ) -> Float
    benefitOverCost ( fid, fd ) =
      ( fromIntegral $ pri $ fd ^. fd_priority )
      * p ^^ ( S.size $ fd ^. fd_seen ) * ( 1 - p )
      / fromIntegral ( max 1 $ fidDiskSpace fid )

    p = 0.2 -- Probability one copy of the file disappears

    pri :: Priority -> Int
    pri ( Priority i ) = 10 ^ i
    pri DefaultPriority = 1
    pri AsManyAsPossible = pri ( Priority 10 )
    pri _ = 0

    byRev :: ( a -> a -> c ) -> ( b -> a ) -> ( b -> b -> c )
    byRev f t = \ a b -> f ( t b ) ( t a )
  

-- | Backup the whole contents of sid:path into dest
backupPath :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k )
              => StorageId -> FilePath -> FilePath -> F4LMonad k m ()
backupPath sid path dest = do
  st <- use $ st_knowledge . kd_contents . to ( M.findWithDefault ST.empty sid )
  fidstore <- use $ st_opEnv . openv_fids
  let fids = map ( fidByIndex' fidstore . snd ) $ subFiles ( splitDirectories path ) $ atTime endOfTime st
  backup fids dest 1000000000000


backup :: ( DelFSMonad m, BufferMonad m, Knowledge k, KnownDataClass k )
          => [ FileId ] -> FilePath -> Int -> F4LMonad k m ()
backup fids dest maxSize = onErrorWarn $ do
  lift $ st_storages .= SM.empty
  fullData <- lift $ use st_knowledge
  envid <- lift $ use st_envid
  say Timid $ "Env: " ++ envid
  storages <- lift storagesForEnv
  createDirectoryIfMissing True dest
  sid <- forceGetStorageId dest
  let seenidxs = S.fromList $ FT.files $ atTime endOfTime $ M.findWithDefault ST.empty sid $ view kd_contents fullData
  say SayAll $ "Seen fid indices: " ++ show seenidxs
  let idxDataMap = view kd_files fullData
  fidstore <- use $ st_opEnv . openv_fids
  let idxData = map ( \ idx -> ( idx, IM.findWithDefault noFileData idx idxDataMap ) ) $ map ( fidIndex' fidstore ) fids
  let newfids = map ( mapFst $ fidByIndex' fidstore ) $ flip filter idxData $ not . ( `S.member` seenidxs ) . fst
  lift $ do
    withStorage sid dest
    runStateT ( backupFids storages maxSize newfids ) (0, 0)
    cleanLine
  lift $ st_storages .= SM.empty
  where
    backupFids :: ( DelFSMonad m, BufferMonad m, Knowledge k ) => SM.Map StorageId [ FilePath ] -> Int -> [(FileId, FileData)] -> StateT (Int, Int) ( F4LMonad k m ) ()
    backupFids storages s fids = -- liftIO ( putStrLn $ "Backing up size: " ++ show s ) >>
      let nfids = dropWhile ( ( > s ) . fidDiskSpace . fst ) fids
      in
       case nfids of
         ( fid, dat ) : rest -> do
           let ( dp, destPath ) = destFidPath fid
           isThere <- doesFileExist destPath
           if isThere then
             do -- File already present, but not recorded
               isOld <- runExceptT $ do
                 s <- fileSize destPath
                 when ( s /= fid_length fid ) noMsg
                 fileFid <- fileId destPath
                 when ( fileFid /= fid ) noMsg
               case isOld of
                 Left _ -> say Quiet ( "Skipping " ++ destPath ++ " present with non-matching FileId" ) >> backupFids storages s rest
                 Right _ -> do
                   say Timid ( "Found " ++ destPath ++ " already present; adding" ) >> backupFids storages s rest
                   lift $ do
                     setFid fid
                     addOp BackupPath
                     addOp Seen
             else
             do
               lift $ onErrorWarn $ createDirectoryIfMissing True dp
               ( nfiles, donesize ) <- get
               let candidates = fileCandidates storages dat
               res <- lift $ runExceptT $ checkedCopyFileCandidates fid candidates destPath $ hShow nfiles ++ " " ++ hShow donesize ++ "\t"
               -- TODO: deal with disk full error
               case res of
                 Right _ -> do
                   _1 += 1
                   _2 += fidDiskSpace fid
                   lift $ do
                     setFid fid
                     addOp BackupPath
                     addOp Seen
                   backupFids storages ( s - fidDiskSpace fid ) rest
                 Left _err -> do
                   say Chatty $ "Skipping on error: " ++ _err
                   backupFids storages s rest
         _ ->
           return ()

    destFidPath fid =
      let ( p, fn ) = fidToBDest fid
      in ( combine dest p, combine dest $ combine p fn )


filesInStorageSubAtTimeSpec :: ( Monad m, KnownDataClass k )
                               => TimeSpec -> StorageId -> FilePath -> StateT (F4LState k ActiveSession) m [([FilePath], FidIndex)]
filesInStorageSubAtTimeSpec timeSpec sid srcpath =
  let
    pfx = splitDirectories srcpath
  in
    use $ st_knowledge . kd_contents . to ( M.findWithDefault ST.empty sid ) . to ( subFiles pfx . atTimeSpec timeSpec )


-- | Fail on errors with DB or destination
--   Report result: file id, destination path, flag
--   flag = Nothing -> file was already present
--   flag = Just True -> file restored
--   flag = Just False -> file not restored
restorePath :: ( DelFSMonad m, BufferMonad m, KnownDataClass k, Knowledge k ) 
  => Bool -> TimeSpec -> StorageId -> FilePath -> FilePath -> ExceptT Error ( F4LMonad k m ) [(FileId, FilePath, Maybe Bool)]
restorePath recorded timeSpec sid srcpath dest = do
  lift $ st_storages .= SM.empty
  createDirectoryIfMissing True dest
  srp <- ( if recorded then forceGetStorage else fakeGetStorage ) dest
  let destsid = rp_sid srp
  let sPath = rp_sPath srp
  let relDest = rp_relPath srp
  lift $ when recorded $ withStorage destsid dest
  genRestore ( tryCopy recorded ) timeSpec sid srcpath sPath relDest


-- | Fail on errors with DB or destination
--   Report result: file id, destination path, flag
--   flag = Nothing -> file was already present
--   flag = Just True -> file restored
--   flag = Just False -> file not restored
symlinkPath :: ( SymlinkFSMonad m, BufferMonad m, KnownDataClass k, Knowledge k ) 
  => TimeSpec -> StorageId -> FilePath -> FilePath -> ExceptT Error ( F4LMonad k m ) [(FileId, FilePath, Maybe Bool)]
symlinkPath timeSpec sid srcpath dest = do
  lift $ st_storages .= SM.empty
  createDirectoryIfMissing True dest
  srp <- fakeGetStorage dest
  let sPath = rp_sPath srp
  let relDest = rp_relPath srp
  genRestore trySymlink timeSpec sid srcpath sPath relDest


genRestore :: ( FSMonad m, BufferMonad m, KnownDataClass k, Knowledge k ) 
           => FileAction k m a -> TimeSpec -> StorageId -> FilePath -> FilePath -> FilePath -> ExceptT Error ( F4LMonad k m ) [a]
genRestore f timeSpec sid srcpath sPath relDest =
  lift $ do
    storages <- storagesForEnv
    toCopy <- filesInStorageSubAtTimeSpec timeSpec sid srcpath
    say Timid $ show $ toCopy
    kdFiles <- use $ st_knowledge . kd_files
    fidstore <- use $ st_opEnv . openv_fids
    res <- mapM ( \ ( path, idx ) -> f storages kdFiles idx ( fidByIndex' fidstore idx ) sPath $ joinPath $ relDest : path ) toCopy
    cleanLine
    st_storages .= SM.empty    
    return res


type FileAction k m a = SM.Map StorageId [FilePath] -> IM.IntMap FileData -> FidIndex -> FileId -> FilePath -> FilePath -> F4LMonad k m a


tryDo :: ( FSMonad m, BufferMonad m, Knowledge k )
         => ( FileId -> [FilePath] -> FilePath -> ExceptT [Char] ( F4LMonad k m ) () )
      -> ( FileId -> FilePath -> F4LMonad k m () )
      -> String -> FileAction k m (FileId, FilePath, Maybe Bool)
tryDo act save label storages files idx fid storagePath relDestPath = do
  let destPath = joinPath [ storagePath, relDestPath ]
  say Timid $ label ++ " " ++ show fid ++ " to " ++ destPath
  outputLineMsg $ label ++ " " ++ show fid ++ " to " ++ destPath
  isThere <- doesFileExist destPath
  if isThere then
    do
      say Chatty "Skipping"
      return ( fid, relDestPath, Nothing )
    else
    do
      res <- runExceptT $ do
        createDirectoryIfMissing True $ takeDirectory destPath
        let mdat = IM.lookup idx files
        case mdat of
          Just dat -> do
            let candidates = fileCandidates storages dat
            done <- lift $ runExceptT $ act fid candidates destPath
            case done of
              Right _ -> do
                lift $ save fid relDestPath
                return ( fid, relDestPath, Just True )
              Left err -> do
                say Chatty $ "Skipping: not found anywhere (" ++ err ++ ")"
                return ( fid, relDestPath, Just False )
          Nothing -> do
            say Chatty "Skipping: not found anywhere"
            return ( fid, relDestPath, Just False )
      return $ either ( const ( fid, relDestPath, Just False ) ) id res


tryCopy :: ( DelFSMonad m, BufferMonad m, Knowledge k )
           => Bool -> FileAction k m (FileId, FilePath, Maybe Bool)
tryCopy recorded = tryDo act save "Copying"
  where
    act fid candidates destPath = checkedCopyFileCandidates fid candidates destPath ""
    save fid relDestPath = when recorded $ do
      setFid fid
      setPath relDestPath
      addOp Seen


trySymlink :: ( SymlinkFSMonad m, BufferMonad m, Knowledge k )
           => FileAction k m (FileId, FilePath, Maybe Bool)
trySymlink = tryDo act ( \ _ _ -> return () ) "Symlinking"
  where
    act fid candidates destPath = symlinkFileCandidates fid candidates destPath ""


checkBackup :: FSMonad m => Bool -> FilePath -> m () -- [ ( FilePath, FileId, String ) ]
checkBackup quick dir = recur
  where
    showMsg ( fn, _fid, msg ) = outputMsg $ joinPath [ dir, fn ] ++ "\t" ++ msg
    recur = do
      l1 <- onErrorUseDefault [] $ getCleanDirectoryContents dir
      mapM_ recur1 $ filter ( (==) 2 . length ) l1
    recur1 l1 = do
      l2 <- onErrorUseDefault [] $ getCleanDirectoryContents $ joinPath [ dir, l1 ]
      mapM_ ( recur2 l1 ) $ filter ( (==) 2 . length ) l2
    recur2 l1 l2 = do
      fn <- onErrorUseDefault [] $ getCleanDirectoryContents $ joinPath [ dir, l1, l2 ]
      mapM_ check [ joinPath [ l1, l2, x ] | x <- fn ]
    check fn = do
      case fidFromBDest fn of
        Nothing ->
          return ()
        Just fid -> do
          let fullPath = joinPath [ dir, fn ]
          s <- onErrorUseDefault (-1) $ fileSize fullPath
          if s /= fid_length fid
            then showMsg ( fn, fid, "Length mismatch" )
            else unless quick $ do
            actualFid <- onErrorUseDefault Nothing $ liftM Just $ fileId fullPath
            case actualFid of
              Nothing -> showMsg ( fn, fid, "Unable to compute file ID" )
              Just x | x /= fid -> showMsg ( fn, fid, "File ID mismatch" )
              _ -> return ()
