{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE DeriveDataTypeable, GeneralizedNewtypeDeriving #-}
module F4L.FSMonad where

import F4L( FileAttr, FilePathAttr( Property ) )
import F4L.Utils( ifM, uuidString )
import qualified F4L.Utils as U( fileSize )

import Control.Concurrent( forkIO )
import Control.Monad
import qualified Control.Monad.Catch as EX
import Control.Monad.Trans.Except( ExceptT(..), mapExceptT, throwE )
import Control.Exception( bracket, tryJust, Exception )
import Control.Monad.State.Strict( liftIO, StateT, evalStateT, get, MonadIO, MonadState, put, runStateT )
import Control.Monad.Trans.Class
import qualified Data.ByteString.Lazy as BL hiding ( hPutStr )
import qualified Data.ByteString as BS hiding ( hPutStr )
import Data.List( sort )
import Data.Maybe( maybeToList, fromMaybe )
import qualified Data.Map as M
import Data.Typeable
import Data.Time.Clock.POSIX( getPOSIXTime )
import Graphics.HsExif( dateTimeOriginal, parseFileExif )
import System.Console.Haskeline.MonadException( MonadException )
import qualified System.Console.Terminal.Size as TSize
import qualified System.Directory as Dir
import System.FilePath( joinPath )
import System.IO( withBinaryFile, hClose, hFileSize, hFlush, hPutStr, openFile, stderr, withFile, Handle, IOMode( ReadMode, AppendMode, WriteMode ) )
import System.IO.Error( isDoesNotExistError )
import System.Posix.Files( createSymbolicLink )


type Error = String


noError = "Unspecified error"


noMsg :: Monad m => ExceptT Error m a
noMsg = throwE noError


data WritingError = WhileOpening Error | WhileWriting Error
                  deriving ( Show, Typeable )


data IsFileOrDirectory = IsFile | IsDirectory


instance EX.Exception WritingError


-- | Monad with filesystem-like operations
class Monad m => FSMonad m where
  outputMsg :: String -> m ()
  outputLineMsg :: String -> m ()
  outputLineMsg _ = return ()
  cleanLine :: m ()
  cleanLine = return ()
  exifProperties :: FilePath -> m [FileAttr]
  generateUuid :: m String
  prompt :: String -> m String
  getNow :: m Int
  readFromFile :: FilePath -> ExceptT Error m String
  actOnLengthAndBLFromFile :: Integral i => FilePath -> ( i -> BL.ByteString -> m a ) -> ExceptT Error m a
  processLengthAndBLFromFile :: Integral i => FilePath -> ( i -> BL.ByteString -> a ) -> ExceptT Error m a
  processLengthAndBLFromFile f act = actOnLengthAndBLFromFile f ( liftM ( liftM return ) act )
  processLengthAndBLFromFile' :: Integral i => FilePath -> ( i -> BL.ByteString -> a ) -> ExceptT Error m a
  -- | Strict version. Read all file and close
  processLengthAndBLFromFile' = processLengthAndBLFromFile
  appendToFile :: FilePath -> String -> ExceptT Error m ()
  appendBLsToFile :: FilePath -> [BL.ByteString] -> ExceptT Error m ()
  -- | The default implementation is inefficient. Better specialize in instances
  writeBLToFileAndDo :: FilePath -> a -> ( BS.ByteString -> a -> m a ) -> BL.ByteString -> ExceptT WritingError m a
  writeBLToFileAndDo f a act bl = do
    mapExceptT ( liftM $ mapLeft WhileWriting ) $ appendBLsToFile f [bl]
    lift $ foldM ( flip act ) a $ BL.toChunks bl
    where
      mapLeft :: ( a -> b ) -> Either a c -> Either b c
      mapLeft f ( Left a ) = Left $ f a
      mapLeft _ ( Right x ) = Right x
  canonicalizePath :: FilePath -> m FilePath
--  copyFile :: FilePath -> FilePath -> ExceptT Error m ()
  doesDirectoryExist :: FilePath -> m Bool
  doesFileExist :: FilePath -> m Bool
  doesFileOrDirectoryExist :: FilePath -> m ( Maybe IsFileOrDirectory )
  doesFileOrDirectoryExist p = do
    ifM ( doesFileExist p ) ( return $ Just IsFile ) (
      ifM ( doesDirectoryExist p ) ( return $ Just IsDirectory ) ( return Nothing ) )
  fileSize :: Integral i => FilePath -> ExceptT Error m i
  getCurrentDirectory :: m FilePath
  getDirectoryContents :: FilePath -> ExceptT Error m [FilePath]
  createDirectoryIfMissing :: Bool -> FilePath -> ExceptT Error m ()
  myCatch :: Exception e => m a -> ( e -> m a ) -> m a


getDirectoryContentsWithType :: FSMonad m => FilePath -> ExceptT Error m [(Bool, FilePath)]
getDirectoryContentsWithType dir = do
  fs <- getDirectoryContents dir
  forM fs $ \f -> do
    isDir <- lift $ doesDirectoryExist $ joinPath [dir, f]
    return (isDir, f)


-- | Directory contents without . and ..
getCleanDirectoryContents dir = 
  getDirectoryContents dir >>= return . filter ( not . flip elem [ ".", ".." ] ) >>= return . sort


-- | FSMonad with deletion capability
class FSMonad m => DelFSMonad m where
  removeFile :: FilePath -> ExceptT Error m ()


-- | FSMonad with parallel execution capability
class FSMonad m => ParFSMonad m where
  fork :: m () -> m ()


-- | FSMonad with symlink capability
class FSMonad m => SymlinkFSMonad m where
  -- | symlink realpath linkpath
  symlink :: FilePath -> FilePath -> ExceptT Error m ()


instance FSMonad m => FSMonad ( StateT s m ) where
  outputMsg = lift . outputMsg
  outputLineMsg = lift . outputLineMsg
  cleanLine = lift cleanLine
  exifProperties = lift . exifProperties
  generateUuid = lift generateUuid
  prompt = lift . prompt
  getNow = lift getNow
  readFromFile = mapExceptT lift . readFromFile
  actOnLengthAndBLFromFile f act = do
    s <- get
    (a, ns) <- mapExceptT lift $ actOnLengthAndBLFromFile f ( \ i dat -> flip runStateT s $ act i dat )
    put ns
    return a
  processLengthAndBLFromFile = liftM ( liftM $ mapExceptT lift ) processLengthAndBLFromFile
  processLengthAndBLFromFile' = liftM ( liftM $ mapExceptT lift ) processLengthAndBLFromFile'
  appendToFile fp s = mapExceptT lift $ appendToFile fp s
  appendBLsToFile = liftM ( liftM $ mapExceptT lift ) appendBLsToFile
  writeBLToFileAndDo p a act bl = do
    s <- get
    (a, ns) <- mapExceptT lift $ writeBLToFileAndDo p ( a, s ) ( unlift act ) bl
    put ns
    return a
    where
      unlift :: ( BS.ByteString -> a -> StateT s m a ) -> BS.ByteString -> ( a, s ) -> m ( a, s )
      unlift f bs (a, s) = flip runStateT s $ f bs a
  canonicalizePath = lift . canonicalizePath
--  copyFile = liftM ( liftM $ mapExceptT lift ) copyFile
  doesDirectoryExist = lift . doesDirectoryExist
  doesFileExist = lift . doesFileExist
  fileSize = mapExceptT lift . fileSize
  getCurrentDirectory = lift getCurrentDirectory
  getDirectoryContents = mapExceptT lift . getDirectoryContents
  createDirectoryIfMissing = liftM ( liftM $ mapExceptT lift ) createDirectoryIfMissing
  myCatch act proc = do
    s <- get
    res <- lift $ myCatch (runStateT act s >>= return . Right) (return . Left)
    case res of
      Right (a, ns) -> do
        put ns
        return a
      Left e ->
        proc e


instance DelFSMonad m => DelFSMonad ( StateT s m ) where
  removeFile = mapExceptT lift . removeFile


instance ParFSMonad m => ParFSMonad ( StateT s m ) where
  fork a = do
    s <- get
    lift $ fork $ evalStateT a s


instance SymlinkFSMonad m => SymlinkFSMonad ( StateT s m ) where
  symlink real link = mapExceptT lift $ symlink real link


ioCatch :: IO a -> ExceptT Error IO a
ioCatch a = ExceptT $ EX.handleIOError ( return . Left . show ) $ liftM Right a


withFile' :: FilePath -> IOMode -> (Handle -> IO a) -> ExceptT WritingError IO a
withFile' name mode = toErr . bracket ( toEx WhileOpening $ openFile name mode ) ( toEx WhileWriting . hClose ) . ( toEx WhileWriting . )
  where
    toErr :: IO a -> ExceptT WritingError IO a
    toErr = ExceptT . EX.try
    toEx :: Exception ex => ( String -> ex ) -> IO a -> IO a
    toEx f a = EX.handleIOError ( EX.throwM . f . show ) a


instance FSMonad IO where
  outputMsg = putStrLn

  outputLineMsg msg = do
    tsize <- ( liftM $ fmap TSize.width ) TSize.size
    hPutStr stderr $ "\r\ESC[K" ++ tabTake 0 ( fromMaybe 80 tsize ) msg
    where
      tabTake _ _ [] = []
      tabTake _ 0 _ = []
      tabTake pos left ( '\t' : xs ) = case 8 - pos `mod` 8 of
        i | i > left -> []
        i -> '\t' : tabTake ( pos + i ) ( left - i ) xs
      tabTake pos left ( x : xs ) = x : tabTake ( pos + 1 ) ( left - 1 ) xs

  cleanLine = outputLineMsg "\r"

  exifProperties path = do
    mexif <- liftIO $ EX.catch ( parseFileExif path >>= return . ( either ( const Nothing ) Just ) ) skipEx
    return $ maybeToList $ fmap ( Property "exif:dateTime" . show ) $ mexif >>= M.lookup dateTimeOriginal
    where
      skipEx :: EX.SomeException -> IO ( Maybe a )
      skipEx _ = return Nothing

  prompt msg = do
    putStrLn msg
    getLine

  generateUuid = uuidString
  
  readFromFile = ioCatch . readFile

  getNow = liftM round getPOSIXTime

  actOnLengthAndBLFromFile f action =
    ioCatch $ withBinaryFile f ReadMode $ \h -> do
      size <- hFileSize h
      dat <- BL.hGetContents h
      res <- action ( fromIntegral size ) dat
      res `seq` return res

  processLengthAndBLFromFile f action = actOnLengthAndBLFromFile f ( \ a b -> return $ action a b )

  processLengthAndBLFromFile' f action =
    ioCatch $ withBinaryFile f ReadMode $ \h -> do
      size <- hFileSize h
      liftM ( action $ fromIntegral size ) $ BL.hGet h $ fromIntegral size

  appendToFile = liftM ( liftM ioCatch ) appendFile
  appendBLsToFile f bls = ioCatch $ withFile f AppendMode $ \ h -> do
    mapM_ ( BL.hPut h ) bls
  -- TODO: when failing, return whether error was on file opening or in writing
  writeBLToFileAndDo f a act bl = withFile' f WriteMode $ \ h -> do
    foldM ( act' h ) a $ BL.toChunks bl
    where act' h a s = do
            BS.hPut h s
            act s a
  canonicalizePath p = do
    cp <- tryJust ( guard . isDoesNotExistError ) $ Dir.canonicalizePath p
    return $ either (const p) id cp
--  copyFile = liftM ( liftM ioCatch ) Dir.copyFile
  doesDirectoryExist = Dir.doesDirectoryExist
  doesFileExist = Dir.doesFileExist
  fileSize = ioCatch . U.fileSize
  getCurrentDirectory = Dir.getCurrentDirectory
  getDirectoryContents = ioCatch . Dir.getDirectoryContents
  createDirectoryIfMissing = liftM ( liftM ioCatch ) Dir.createDirectoryIfMissing
  myCatch = EX.catch


instance DelFSMonad IO where
  removeFile = ioCatch . Dir.removeFile


instance ParFSMonad IO where
  fork a = void $ forkIO a


instance SymlinkFSMonad IO where
  -- TODO: use createFileLink from directory>=1.3.1.0
  symlink real link = ioCatch $ createSymbolicLink real link


-- | Same as IO, but record operations to a log file
newtype LoggingIO a = LoggingIO { deLoggingIO :: StateT ( Maybe Handle ) IO a }
                    deriving ( Applicative, Functor, Monad, MonadIO, DelFSMonad, ParFSMonad, SymlinkFSMonad, MonadException, MonadState ( Maybe Handle ) )


loggingHandle :: LoggingIO Handle
loggingHandle = do
  old <- get
  case old of
       Just h -> return h
       Nothing -> do
         h <- liftIO $ openFile "/tmp/f4l-logging.log" AppendMode
         put $ Just h
         return h


loggingLog :: Show a => String -> a -> LoggingIO ()
loggingLog label s = loggingHandle >>= \ h -> ( liftIO $ hPutStr h ( label ++ ": " ++ show s ++ "\n") >> hFlush h )


loggingLog' s a = lift $ loggingLog s a


-- | Log single argument
loggingLogO label f1 = \ a -> loggingLog' label a >> f1 a


-- | Log two arguments
loggingLogOO label f2 = \ a b -> loggingLog' label (a, b) >> f2 a b


-- | Log first of two arguments
loggingLogOX label f2 = \ a b -> loggingLog' label a >> f2 a b


-- | Log first of four arguments
loggingLogOXXX label f4 = \ a b c d -> loggingLog label a >> f4 a b c d


instance FSMonad LoggingIO where
  outputMsg = LoggingIO . outputMsg
  outputLineMsg = LoggingIO . outputLineMsg
  cleanLine = LoggingIO cleanLine
  exifProperties = LoggingIO . exifProperties
  generateUuid = LoggingIO generateUuid
  prompt = LoggingIO . prompt
  getNow = LoggingIO getNow
  readFromFile = loggingLogO "readFromFile" $ mapExceptT LoggingIO . readFromFile
  actOnLengthAndBLFromFile f act = do
    loggingLog' "actOnLengthAndBLFromFile" f
    mapExceptT LoggingIO $ actOnLengthAndBLFromFile f $ liftM ( liftM deLoggingIO ) act
  processLengthAndBLFromFile = loggingLogOX "processLengthAndBLFromFile" $ liftM ( liftM $ mapExceptT LoggingIO ) processLengthAndBLFromFile
  processLengthAndBLFromFile' = loggingLogOX "processLengthAndBLFromFile'" $ liftM ( liftM $ mapExceptT LoggingIO ) processLengthAndBLFromFile'
  appendToFile fp s = do
    loggingLog' "appendToFile" fp 
    mapExceptT LoggingIO $ appendToFile fp s
  appendBLsToFile = loggingLogOX "appendBLsToFile" $ liftM ( liftM $ mapExceptT LoggingIO ) appendBLsToFile
  writeBLToFileAndDo p a act bl = do
    loggingLog' "writeBLToFileAndDo" p
    mapExceptT LoggingIO $ writeBLToFileAndDo p a ( liftM ( liftM deLoggingIO ) act ) bl
  canonicalizePath = LoggingIO . canonicalizePath
  doesDirectoryExist d = loggingLog "doesDirectoryExist" d >> LoggingIO ( doesDirectoryExist d )
  doesFileExist f = do
    loggingLog "doesFileExist" f
    LoggingIO $ doesFileExist f
  doesFileOrDirectoryExist f = do
    loggingLog "doesFileOrDirectoryExist" f
    LoggingIO $ doesFileOrDirectoryExist f
  fileSize = loggingLogO "fileSize" $ mapExceptT LoggingIO . fileSize
  getCurrentDirectory = loggingLog "getCurrentDirectory" () >> LoggingIO getCurrentDirectory
  getDirectoryContents = loggingLogO "getDirectoryContents" $ mapExceptT LoggingIO . getDirectoryContents
  createDirectoryIfMissing = loggingLogOO "createDirectoryIfMissing" $ liftM ( liftM $ mapExceptT LoggingIO ) createDirectoryIfMissing
  myCatch act proc = LoggingIO $ myCatch ( deLoggingIO act ) ( liftM deLoggingIO proc )
