{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE DeriveGeneric, DeriveAnyClass #-}

module F4L.FidStore (
    FidIndex
  , FidStore
  , empty
  , fidIndex
  , fidIndex'
  , fidByIndex
  , fidByIndex'
  , findShorthand
  , findFid
  , findFidsByStringHash
  , withFid
  ) where

import F4L
import F4L.Utils( fromJust', getList, getStrict, hashFromS, hashToS )

import Control.DeepSeq
import Control.Monad( liftM, mzero, msum )
import qualified Data.Binary as B
import Data.Binary( Binary )
import qualified Data.ByteString as BS
import qualified Data.ByteString.Short as SBS
import qualified Data.IntMap.Strict as IM
import Lens.Micro.Platform
import Data.List( find, inits, isPrefixOf )
import Data.Map.Strict( Map )
import qualified Data.Map.Strict as M
import qualified Data.Store as DS
import GHC.Generics( Generic )


type FidIndex = Int


-- | Keeps track of known FileIds
data FidStore = FidStore {
    _fs_by_short :: !( Map SBS.ShortByteString FidIndex )
  , _fs_idx :: !( IM.IntMap FileId )
  , _fs_size :: !Int
  }
  deriving (Eq, Show, Generic, DS.Store)


instance NFData FidStore where
  rnf ( FidStore a b c ) = rnf a `seq` rnf b `seq` rnf c


instance Binary FidStore where
  put ( FidStore a b c ) = B.put ( M.toList a ) >> B.put b >> B.put c
  get = do
    a <- liftM M.fromList $ getList getStrict
    b <- liftM IM.fromList $ getList getStrict
    c <- B.get
    return $ FidStore a b c


empty = FidStore M.empty IM.empty 0


fs_by_short :: Lens' FidStore ( Map SBS.ShortByteString Int )
fs_by_short = lens _fs_by_short $ \ s a -> s { _fs_by_short = a }
{-# INLINE fs_by_short #-}


fs_idx :: Lens' FidStore ( IM.IntMap FileId )
fs_idx = lens _fs_idx $ \ s a -> s { _fs_idx = a }
{-# INLINE fs_idx #-}


fs_size :: Lens' FidStore Int
fs_size = lens _fs_size $ \ s a -> s { _fs_size = a }
{-# INLINE fs_size #-}


findShorthand :: FidStore -> FileId -> Maybe SBS.ShortByteString
findShorthand fids fid @ ( FileId _ h ) = find cond $ map SBS.toShort $ drop 1 $ BS.inits $ SBS.fromShort h
  where
    cond short = ( M.lookup short ( view fs_by_short fids ) >>= fidByIndex fids ) == Just fid


-- | Returns the matching fid and index (if unique)
findFidIdx :: FidStore -> SBS.ShortByteString -> Maybe ( FileId, FidIndex )
findFidIdx fids short = msum $ map cond $ reverse $ BS.inits short'
  where
    short' = SBS.fromShort short
    cond sh = do
      idx <- M.lookup ( SBS.toShort sh ) ( view fs_by_short fids )
      x @ ( FileId _ h ) <- view ( fs_idx . at idx ) fids
      if short' `BS.isPrefixOf` ( SBS.fromShort h )
        then return (x, idx)
        else mzero


findFid :: FidStore -> SBS.ShortByteString -> Maybe FileId
findFid fids = fmap fst . findFidIdx fids


fidIndex :: FidStore -> FileId -> Maybe FidIndex
fidIndex fids f = fmap snd $ findFidIdx fids ( fid_hash f )


fidIndex' :: FidStore -> FileId -> FidIndex
fidIndex' fids fid = fromJust' ( "Unknown FileId " ++ show fid ) $ fidIndex fids fid


fidByIndex :: FidStore -> FidIndex -> Maybe FileId
fidByIndex fids i = IM.lookup i $ view fs_idx fids


fidByIndex' :: FidStore -> FidIndex -> FileId
fidByIndex' fids i = fromJust' ( "Unknown FidIndex " ++ show i ) $ fidByIndex fids i


findFidsByStringHash :: String -> FidStore -> [ FileId ]
findFidsByStringHash short fids =
  let
    pfx = fromJust' ( "Unable to get partial hash from " ++ short ) $ msum $ map hashFromS $ reverse $ inits short
  in
   pfx `seq` filter ( ( short `isPrefixOf` ) . hashToS . fid_hash ) $ IM.elems $ view fs_idx fids


withFid :: FileId -> FidStore -> FidStore
withFid fid @ ( FileId _ h ) fids = withFid' ( drop 1 $ BS.inits ( SBS.fromShort h ) )
  where
    withFid' [] = error $ "Exausted hash length while looking for slot matching " ++ show h
    withFid' ( x : xs ) =
      let sh = SBS.toShort x in
        case M.lookup sh ( view fs_by_short fids ) of
          Nothing ->
            let
              new_size = view fs_size fids + 1
            in
              set fs_size new_size
                $ over fs_idx ( IM.insert new_size fid )
                $ over fs_by_short ( M.insert sh new_size )
                $ fids
          Just idx ->
            case fidByIndex' fids idx of
              fid' | fid == fid' -> fids
              ( FileId _ h' ) | h == h' -> 
                                     error $ "Found a SHA1 collision for " ++ show fid ++ "?"
              _ -> withFid' xs
