{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE FlexibleContexts #-}
module F4L.Interactive where

import F4L
import F4L.Actions
import F4L.BufferMonad
import F4L.FidStore
import F4L.FileTree as FT
import F4L.FSMonad
import F4L.Interactive.IParser
import F4L.Knowledge
import F4L.Op( findFidsByStringHash, openv_fids )
import F4L.StoryTree as ST
import F4L.Utils hiding( fileSize )

import Lens.Micro.Platform( at, lens, to, use, (.=), (^.), (.~), Getting, Lens' )
import Control.Monad.Trans.Except( runExceptT )
import Control.Monad.State.Strict
import Control.Monad.Trans.Maybe
import Control.DeepSeq( force )
import Data.Function( on )
import Data.List( isPrefixOf, sort, sortBy, foldl', intercalate )
import Data.Map.Strict( Map )
import Data.Maybe( fromMaybe, mapMaybe, listToMaybe )
import qualified Data.Map.Strict as SM
import qualified Data.Set as S
import Data.Tree( drawTree )
import Text.Parsec
import System.Console.Haskeline
import System.FilePath( joinPath, splitFileName, addTrailingPathSeparator, splitDirectories )
import System.IO.Unsafe( unsafePerformIO )


data InteractionContext = Local | InStorage StorageId
                        deriving (Eq, Ord)


-- | Enhanced Knowledge used when performing operations within an interactive session
type InteractionKnowledge =
  ( KnownData
  , ( SeenStorages
    , ( SeenEnvironments
      , ( FactCounter
        , FileSet
        ) ) ) )


icToSid Local = Nothing
icToSid ( InStorage sid ) = Just sid


-- | Structure containing the state built-up during an interactive session
--   containing options etc.
data InteractionStatus = IStatus {
    _is_context :: InteractionContext
  , _is_prev_context :: InteractionContext
  , _is_time :: TimeSpec
  , _is_paths :: Map InteractionContext FilePath
  , _is_quit :: Bool
  , _is_commit :: Bool
  , _is_tags :: [String]
  , _is_props :: [(String, String)]
  }

initialInteractionStatus :: FSMonad m => m InteractionStatus
initialInteractionStatus = do
  cwd <- getCurrentDirectory
  return $ IStatus Local Local ( At endOfTime ) (SM.singleton Local cwd) False False [] []

-- is_envid = lens _is_envid $ \ s a -> s { _is_envid = a }

is_context :: Lens' InteractionStatus InteractionContext
is_context = lens _is_context $ \ s a -> s { _is_context = a, _is_prev_context = _is_context s }

-- is_prev_context :: Getter InteractionStatus InteractionContext
is_prev_context :: Getting a InteractionStatus InteractionContext
is_prev_context = to _is_prev_context

is_time :: Lens' InteractionStatus TimeSpec
is_time = lens _is_time $ \ s a -> s { _is_time = a }

is_sid = is_context . to icToSid

is_paths :: Lens' InteractionStatus (Map InteractionContext FilePath)
is_paths = lens _is_paths ( \ s a -> s { _is_paths = a } )

is_path :: Lens' InteractionStatus FilePath
is_path = lens v s
  where
    v x = SM.findWithDefault "/" (_is_context x) (_is_paths x)
    s x a = x { _is_paths = SM.insert (_is_context x) a ( _is_paths x ) }

is_local_path :: Lens' InteractionStatus FilePath
is_local_path = lens v s
  where
    v x = SM.findWithDefault "/" Local (_is_paths x)
    s x a = x { _is_paths = SM.insert Local a ( _is_paths x ) }

{-
is_data :: Lens' InteractionStatus KnownData
is_data = lens _is_data ( \ s a -> s { _is_data = a } )
-}

is_quit :: Lens' InteractionStatus Bool
is_quit = lens _is_quit ( \ s a -> s { _is_quit = a } )

is_commit :: Lens' InteractionStatus Bool
is_commit = lens _is_commit ( \ s a -> s { _is_commit = a } )


byContext :: Monad m => Interaction m a -> ( StorageId -> Interaction m a ) -> Interaction m a
byContext whenLocal whenStorage =
  use is_sid >>= maybe whenLocal whenStorage


joinDirCanonical :: FSMonad m => String -> String -> m String
joinDirCanonical parent child = canonicalizePath $ joinPath [ parent, child ]

joinDir :: String -> String -> String
joinDir parent child = joinPath $ j [] $ filter ( /= "." ) $ splitDirectories parent ++ splitDirectories child
    where
      j pre ( "/" : ".." : rest ) = j pre ( "/" : rest )
      j pre ( _x : ".." : rest ) = j ( pre ++ rest ) []
      j pre [] = pre
      j pre ( x : rest ) = j ( pre ++ [ x ] ) rest


type Interaction m = StateT InteractionStatus m


sidContents :: ( Monad m, KnownDataClass k ) => TimeSpec -> StorageId -> [FilePath] -> F4LMonad k m (Maybe (FileTree FidIndex))
sidContents ts sid path = do
  contents <- use $ st_knowledge . kd_contents . to ( atTimeSpec ts . SM.findWithDefault ST.empty sid )
  return $ findNode path contents


dummyDir = fst $ splitFileName "abc" -- TODO: de-uglify

fileComplete f = (False, Completion f f True)

dirComplete d = let td = addTrailingPathSeparator d in (True, Completion td td False)


listLocalPaths, listCurrentPaths ::
  FSMonad m
  => String -> Interaction ( F4LMonad InteractionKnowledge m ) [(Bool, Completion)]
listLocalPaths pre = do
  cwd <- use is_path
  let (dir, file) = splitFileName pre
  do
    let rdir = joinPath [cwd, dir]
    dc <- onErrorUseDefault [] ( getDirectoryContents rdir ) >>= return . filter ( file `isPrefixOf` )
    forM dc $ \ f -> do
      let p = joinPath $ [ d | d <- [dir], d /= dummyDir ] ++ [ f ]
      isDir <- doesDirectoryExist $ joinPath [cwd, p]
      return $ if isDir then dirComplete p else fileComplete p

listCurrentPaths pre = do
  cwd <- use is_path
  let (dir, file) = splitFileName pre
  byContext ( listLocalPaths pre ) $ \ sid -> do
      let rdir = dropWhile ( == "/" ) $ splitDirectories $ joinDir cwd dir
      ts <- use is_time
      node <- lift $ sidContents ts sid rdir
      case node of
        Just ( Dir st ) ->
          return $ do -- list monad
            (subPath, node) <- map ( mapFst shortToPath ) $ SM.toList st
            let p = joinPath $ [ d | d <- [dir], d /= dummyDir ] ++ [subPath ]
            if file `isPrefixOf` subPath then (
               case node of
                 File _ -> return $ fileComplete p
                 Dir _ -> return $ dirComplete p
               ) else []
        _ -> return []


getFidStore :: Monad m
  => Interaction ( F4LMonad InteractionKnowledge m ) FidStore
getFidStore = lift $ use $ st_opEnv . openv_fids


iComplete :: (MonadException m, DelFSMonad m, SymlinkFSMonad m, BufferMonad m)
             => CompletionFunc ( Interaction ( F4LMonad InteractionKnowledge m ) )
iComplete (rline, _) = do
  parsed <- runParserT (partialParser' parseCmd) () "" $ reverse rline
  -- outputMsg $ show parsed
  let posParses = either (const []) id parsed
  let pos = foldr ( max . fst ) 0 posParses
  let part = drop pos $ reverse rline
  -- Only take maximal parses
  let parses = map snd $ filter ( (== pos) . fst ) posParses
  storages <- lift $ use $ st_knowledge . seen_storages
  environments <- lift $ use $ st_knowledge . seen_environments
  fidStore <- getFidStore
  currContext <- use is_context
  timestamps <- case currContext of
    InStorage s -> lift $ use $ st_knowledge . kd_contents . at s . to ( maybe [] allTimes )
    _ -> return []
  let local_ws = listLocalPaths $ unescape part
  let current_ws = listCurrentPaths $ unescape part
  parses' <- withLazy local_ws $ withLazy ( lift current_ws ) $ mapM ( getComp storages environments fidStore timestamps part ) $ sNub $ sort parses
  return ( reverse $ take pos $ reverse rline
         , nubSort $ join parses' )
  where
    partialParser' :: Stream String m Char => IParser u m (m ()) -> ParsecT String u m [(Int, Complete)]
    partialParser' = partialParser

    nubSort :: [Completion] -> [Completion]
    nubSort = sNubBy sameReplacement . sortBy cmpCompletion

    sameReplacement = (==) `on` replacement

    cmpCompletion = ( compare `on` replacement ) `mappend` ( compare `on` isFinished )

    getFiniteComp part ss =
      [ Completion a ( maybe b ( ( b ++ "\t" ) ++ ) c ) True
      | ( a, b, c ) <- ss
      , part `isPrefixOf` a ]

    escapedCompletion :: String -> Completion -> Maybe Completion
    escapedCompletion part c = liftM ( \ e -> c { replacement = e } ) $ isEscapedPrefix part ( replacement c )

    getComp _ _ _ _ part (CompleteFixed ss) = return $ getFiniteComp part $ map ( \ (a, b) -> (a, a, b) ) ss
    getComp _ _ _ _ part CompleteDir = do
      ws <- getLazy
      return $ mapMaybe ( escapedCompletion part . snd ) $ filter fst ws
    getComp _ _ _ _ part CompleteLocalDir = do
      ws <- lift getLazy
      return $ mapMaybe ( escapedCompletion part . snd ) $ filter fst ws
    getComp _ _ _ _ part CompletePath = do
      ws <- getLazy
      return $ mapMaybe ( escapedCompletion part . snd ) ws
    getComp _ _ _ _ part CompleteLocalPath = do
      ws <- lift getLazy
      return $ mapMaybe ( escapedCompletion part . snd ) ws
    getComp storages _ _ _ part CompleteSid =
      return $ getFiniteComp part [ (k, abbrSid k, listToMaybe v) | (k, v) <- SM.toList storages ]
    getComp _ _ fidStore _ part CompleteFid =
      return $ getFiniteComp part [ (hashToS $ fid_hash fid, hashToS $ fid_hash fid, Just("length " ++ show (fid_length fid) )) | fid <- findFidsByStringHash part fidStore ]
    getComp _ environments _ _ part CompleteEnvid =
      return $ getFiniteComp part [ (k, k, listToMaybe v) | (k, v) <- SM.toList environments ]
    getComp _ _ _ timestamps part CompleteTimeStamp =
      return $ getFiniteComp part $ ("latest", "latest", Nothing) : [ (k, k, Nothing) | k <- map timestampToLocaltime timestamps ]


ls :: ( FSMonad m, KnownDataClass k ) => String -> Interaction (F4LMonad k m) ()
ls rel = do
  dir <- byContext (
    do
      path <- use is_path >>= lift . flip joinDirCanonical rel
      lift $ onErrorWarnWithDefault [] $ getDirectoryContentsWithType path
    ) ( \ sid -> do
           path <- use is_path >>= return . dropWhile (== "/") . splitDirectories . flip joinDir rel
           ts <- use is_time
           contents <- lift $ sidContents ts sid path
           case contents of
             Just ( Dir st ) -> return $ map to2ple $ SM.toList st
             Just ( File _ ) -> return [( False, joinPath path )]
             _ -> return []
      )
  -- TODO: show whether file or dir
  mapM_ outputMsg $ sort $ map toRep dir
    where
      to2ple ( path, Dir _ ) = (True, shortToPath path)
      to2ple ( path, File _ ) = (False, shortToPath path)
      toRep ( False, path ) = path
      toRep ( True, path ) = path ++ "/"


-- | change relative path to path relative to the storage's /
storagePath :: Monad m => FilePath -> Interaction ( F4LMonad InteractionKnowledge m ) FilePath
storagePath p = use is_path >>= return . dropWhile (== '/') . flip joinDir p


iRestore :: ( BufferMonad m, DelFSMonad m )
            => Bool -> StorageId -> FilePath -> FilePath -> Interaction ( F4LMonad InteractionKnowledge m ) ()
iRestore record sid p lp = do
  path <- storagePath p
  lpath <- use is_local_path >>= lift . flip joinDirCanonical lp
  timeSpec <- use is_time
  eres <- lift $ runExceptT $ restorePath record timeSpec sid path lpath
  case eres of
    Left err ->
      outputMsg $ "Error: " ++ err
    Right res -> do
      let acc (_, _, Nothing) (i, j, k) = (i+1, j, k)
          acc (_, _, Just False) (i, j, k) = (i, j+1, k)
          acc (_, _, Just True) (i, j, k) = (i, j, k+1)
      let (skipped, notFound, restored) = foldr acc (0, 0, 0) res
      outputMsg $ show restored ++ " files restored"
      outputMsg $ show skipped ++ " files already present"
      outputMsg $ show notFound ++ " files not found"


iSymlink :: ( BufferMonad m, SymlinkFSMonad m )
            => StorageId -> FilePath -> FilePath -> Interaction ( F4LMonad InteractionKnowledge m ) ()
iSymlink sid p lp = do
  path <- storagePath p
  lpath <- use is_local_path >>= lift . flip joinDirCanonical lp
  timeSpec <- use is_time
  eres <- lift $ runExceptT $ symlinkPath timeSpec sid path lpath
  case eres of
    Left err ->
      outputMsg $ "Error: " ++ err
    Right res -> do
      let acc (_, _, Nothing) (i, j, k) = (i+1, j, k)
          acc (_, _, Just False) (i, j, k) = (i, j+1, k)
          acc (_, _, Just True) (i, j, k) = (i, j, k+1)
      let (skipped, notFound, symlinked) = foldr acc (0, 0, 0) res
      outputMsg $ show symlinked ++ " files symlinked"
      outputMsg $ show skipped ++ " files already present"
      outputMsg $ show notFound ++ " files not found"


cd :: FSMonad m => String -> Interaction m ()
cd newdir = byContext (
  do
    newPath <- use is_path >>= lift . flip joinDirCanonical newdir
    whenM ( lift $ doesDirectoryExist newPath ) $ is_path .= newPath
  ) $ const (
  do
    newPath <- use is_path >>= return . flip joinDir newdir
    is_path .= newPath
  )


showSettings :: ( FSMonad m, KnownDataClass k ) => Interaction ( F4LMonad k m ) ()
showSettings = do
  withRescan <- lift $ use st_withRescan
  lift $ outputMsg $ "Rescan: " ++ show withRescan
  verbosity <- lift $ use st_verbosity
  lift $ outputMsg $ "Verbosity level: " ++ show ( verbosity ^. verb_level )
  lift $ outputMsg $ "Tracing: " ++ show ( verbosity ^. verb_tracing )


showSessions :: FSMonad m => Interaction ( F4LMonad k m ) ()
showSessions = lift $ do
  outputMsg "Warning: uncommited data will not appear"
  onErrorWarn $ fromDB ( return . sessions ) >>= mapM_ ( lift . outputMsg . formatSession . fst )


abbrSid :: StorageId -> String
abbrSid sid = take 6 sid ++ "..."


abbrFid :: FileId -> String
abbrFid fid = take 8 ( hashToS $ fid_hash fid ) ++ "..." ++ ":" ++ show ( fid_length fid )


showStorages :: ( SeenStoragesClass k, FSMonad m ) => Interaction ( F4LMonad k m ) ()
showStorages = lift $ do
  ss <- use $ st_knowledge . seen_storages
  mapM_ s' $ sort $ SM.toList ss
  where
    s' ( sid, x : _ ) = outputMsg $ abbrSid sid ++ " " ++ x
    s' ( sid, _ ) = outputMsg $ abbrSid sid


showUnreachable :: ( KnownDataClass k, FSMonad m ) => Interaction ( F4LMonad k m ) ()
showUnreachable = lift $ do
  (fids, storages) <- unreachable
  unless ( null fids ) $ do
    outputMsg "File IDs currently unreachable:"
    forM_ fids $ outputMsg . show
  unless ( null storages ) $ do
    outputMsg "Previous files are contained in storages:"
    forM_ storages $ outputMsg . abbrSid


showKSize :: ( Knowledge k, FSMonad m ) => Interaction ( F4LMonad k m ) ()
showKSize = lift $ do
  k <- use $ st_knowledge
  let res = unsafePerformIO $ sizeInfo k
  forM_ res $ outputMsg . show


changeToStorage :: ( SeenStoragesClass k, FSMonad m ) => StorageId -> Interaction ( F4LMonad k m ) ()
changeToStorage newsid = do
  ss <- lift $ use $ st_knowledge . seen_storages
  case filter ( newsid `isPrefixOf` ) $ SM.keys ss of
    [ sid ] -> is_context .= InStorage sid
    [] -> outputMsg $ "No storage id starting with " ++ newsid
    x -> outputMsg $ show ( length x ) ++ " storages starting with " ++ newsid ++ ". Use a more specific id"


infoPath :: ( FSMonad m, BufferMonad m )
            => StorageId -> FilePath -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoPath sid p = do
  path <- storagePath p
  ts <- use is_time
  runMaybeT $ do
    tree <- MaybeT $ lift $ use $ st_knowledge . kd_contents . at sid
    node <- MaybeT $ return $ findNode ( splitDirectories path ) $ atTimeSpec ts tree
    fidStore <- lift getFidStore
    case node of
      File fix -> lift $ infoFid $ fidByIndex' fidStore fix
      dir @ ( Dir _ ) -> lift $ infoFids ( Just sid ) [( joinPath ( p : subpath ), fidByIndex' fidStore fix ) | (subpath, fix) <- allFiles dir]
  return ()


readTree :: ( FSMonad m, VerboseMonad m ) => FilePath -> m (FileTree FileId)
readTree p = do
  fileOrDir <- doesFileOrDirectoryExist p
  case fileOrDir of
    Just IsFile ->
      onErrorWarnWithDefault FT.empty $ liftM File $ fileId p
    Just IsDirectory ->
      do
        subPaths <- onErrorWarnWithDefault [] ( getCleanDirectoryContents p )
        foldM ( \ t s -> do
                  st <- readTree ( joinPath [ p, s ] )
                  return $ setPath ( pathToShort s ) st t
              ) FT.empty subPaths
    Nothing -> do
      outputLineMsg $ "File " ++ p ++ " not found"
      return FT.empty
    where
      setPath shortPath toset ( Dir sm ) = Dir ( SM.insert shortPath toset sm )
      setPath _ _ a = a


infoLocalPath :: ( FSMonad m, BufferMonad m )
            => FilePath -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoLocalPath p = do
  path <- use is_local_path >>= lift . flip joinDirCanonical p
  fileOrDir <- doesFileOrDirectoryExist path
  case fileOrDir of
    Just IsFile -> do
      mfid <- onErrorWarnNothing $ fileId path
      case mfid of
        Just fid -> infoFid fid
        Nothing -> return ()
    Just IsDirectory -> do
      t <- readTree path
      infoFids Nothing $ map ( mapFst joinPath ) $ subFiles [] t
    _ -> outputMsg $ p ++ " not found"


infoFids :: ( FSMonad m, BufferMonad m )
           => Maybe StorageId -> [(FilePath, FileId)] -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoFids _msid path_fids = do
  -- TODO: produce more interesting results, coverage etc.
  fidStore <- getFidStore
  let fids = map snd path_fids
  mapM_ infoFid fids
  outputMsg "Summary:"
  -- TODO: use fds <- lift $ mapM ( \ fid -> use $ st_knowledge . kd_files . at' fid ) fids
  coverage <- lift $ fidsCoverage ( fidIndex fidStore ) fids
  outputMsg $ show coverage
  return ()


infoFid :: ( FSMonad m, BufferMonad m )
           => FileId -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoFid fid = do
  outputMsg $ show fid
  fidStore <- getFidStore
  let mfix = fidIndex fidStore fid
  case mfix of
    Nothing ->
      outputMsg $ "No information on " ++ show fid

    Just fix -> do
      candidates <- lift $ findLocalFileCandidates fix
      localPaths <- filterM ( liftM ( == fid_length fid ) . onErrorUseDefault (-1) . fileSize ) candidates
      unless ( null localPaths ) $ do
        outputMsg "Locally at:"
        mapM_ outputMsg localPaths
      mfdat <- lift $ use $ st_knowledge . kd_files . at'' fix
      case mfdat of
        Just fdat -> do
          outputMsg $ "Priority: " ++ show ( fdat ^. fd_priority )
          outputMsg $ "Attrs: " ++ intercalate " - " ( map show $ fdat ^. fd_attrs )
          forM_ ( sort $ map ( mapSnd shortToPath ) $ S.toList $ fdat ^. fd_seen ) $ \ ( sid, path ) -> do
            outputMsg $ "Seen in " ++ sid ++ "\t" ++ path
        Nothing -> return ()


infoFidByHash :: ( FSMonad m, BufferMonad m )
           => String -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoFidByHash h = do
  fidStore <- getFidStore
  mapM_ infoFid $ findFidsByStringHash h fidStore


infoStorage :: FSMonad m => StorageId -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoStorage sid = do
  ts <- use is_time
  fidStore <- getFidStore
  lift $ do
    outputMsg $ "Info for storage " ++ sid
    seen <- use $ st_knowledge . seen_storages . at' sid
    mapM outputMsg $ fromMaybe [] seen
    ss <- use $ st_knowledge . kd_storages . to ( SM.findWithDefault [] sid )
    forM_ ss $ \ ( envid, path ) -> outputMsg ( "Seen on env " ++ envid ++ " in " ++ path )
    envid <- use st_envid
    let seenLocally = map snd $ filter ( (== envid) . fst ) ss
    findStorage sid seenLocally >>= mapM_ ( outputMsg . ( "Currently on " ++ ) )
    st <- use $ st_knowledge . kd_contents . to ( SM.findWithDefault ST.empty sid )
    let (n, l) = foldl' ( \ (n, l) fid -> ( force $ n + 1, force $ l + fid_length fid ) ) (0, 0) $ map ( fidByIndex' fidStore ) $ files $ atTimeSpec ts st
    outputMsg $ hShow n ++ " files"
    outputMsg $ hShow l ++ " bytes"


infoEnvironment :: FSMonad m => EnvironmentId -> Interaction ( F4LMonad InteractionKnowledge m ) ()
infoEnvironment _envid = return () -- To be implemented


showStats :: ( FSMonad m, BufferMonad m ) => F4LMonad InteractionKnowledge m ()
showStats = do
  s <- stats
  sequence_ $ do
    ( label, dat ) <- s
    return $ lift $ outputMsg $ label ++ "\t" ++ dat


debug :: FSMonad m => Interaction ( F4LMonad InteractionKnowledge m ) ()
debug = do
  contents <- lift $ use $ st_knowledge . kd_contents
  ts <- use is_time
  outputMsg $ show $ length $ show $ concatMap ( allFiles . atTimeSpec ts ) $ SM.elems contents
  _files <- lift $ use $ st_knowledge . kd_files
  seenTrees <- lift $ use $ st_knowledge . kd_contents
  paths <- listCurrentPaths ""
  forM_ ( SM.map ( drawTree . toTree ) seenTrees ) outputMsg
  outputMsg $ show $ paths


parseCmd :: ( Stream String n Char, DelFSMonad m, SymlinkFSMonad m, BufferMonad m )
            => IParser u (Interaction n) ( Interaction ( F4LMonad InteractionKnowledge m ) () )
parseCmd = foldr1 orElse $ map terminated [
    someString "quit" `andDo` const ( is_quit .= True )
  , someString "commit" `andDo` const ( is_commit .= True )
  , ( someString "ls" `sp` somePath ) `andDo` ls
  , someString "ls" `andDo` const ( ls "" )
  , ( someString "cd" `sp` someDir ) `andDo` cd
  , ( csHelp `sp` someStringHelp "-" "previous storage" ) `andDo` const ( use is_prev_context >>= (.=) is_context )
  , ( csHelp `sp` someSid ) `andDo` changeToStorage
  , csHelp `andDo` const ( is_context .= Local )
  , inStorage $ const $ ctHelp `sp` someStringHelp "*" "all times merged together" `andDo` const ( (.=) is_time Merged )
  , inStorage $ const $ ctHelp `sp` someTimeStamp `andDo` ( (.=) is_time . At )
  , someString "scan" `sp` someString "--priority" `sp` liftM Just somePriority `sp'` scan'
  , someString "scan" `sp` scan' Nothing
  , someString "show" `sp` someString "coverage" `andDo` const ( lift $ coverage >>= mapM_ ( outputMsg . show ) )
  , someString "show" `sp` someString "stats" `andDo` const ( lift showStats )
  , someString "show" `sp` someString "settings" `andDo` const showSettings
  , someString "show" `sp` someString "sessions" `andDo` const showSessions
  , someString "show" `sp` someString "storages" `andDo` const showStorages
  , someString "show" `sp` someString "unreachable" `andDo` const showUnreachable
  , someString "show" `sp` someString "knowledgeSize" `andDo` const showKSize
  , inStorage $ \ sid -> someString "info" `sp` someString "--file" `sp` somePath `andDo` infoPath sid
  , inLocal $ someString "info" `sp` someString "--file" `sp` somePath `andDo` infoLocalPath
  , inLocal $ someString "info" `sp` someString "--fid" `sp` someFid `andDo` infoFidByHash
  , someString "info" `sp` someString "--storage" `sp` someSid `andDo` infoStorage
  , inStorage $ \ sid -> someString "info" `sp` someString "--storage" `andDo` const ( infoStorage sid )
  , someString "info" `sp` someString "--env" `andDo` const ( lift ( use st_envid ) >>= infoEnvironment )
  , someString "info" `sp` someString "--env" `sp` someEnvid `andDo` infoEnvironment
  , inStorage $ \ sid -> restoreHelp `sp` someString "--track" `sp` somePath `sp'` \ p -> someLocalPath >>= \ lp -> return ( iRestore True sid p lp )
  , inStorage $ \ sid -> restoreHelp `sp` someString "--notrack" `sp` somePath `sp'` \ p -> someLocalPath >>= \ lp -> return ( iRestore False sid p lp )
  , inStorage $ \ sid -> symlinkHelp `sp` somePath `sp'` \ p -> someLocalPath >>= \ lp -> return ( iSymlink sid p lp )
  , backupHelp
    `sp` someString "-s"
    `sp` someSize
    `sp'` \ s -> someLocalDir `andDo` ( lift . flip autoBackup s )
  , inStorage $ \ sid -> backupHelp
    `sp` somePath
    `sp'` \ s -> someLocalDir `andDo` backupStorage sid s
  , inLocal $ backupHelp
    `sp` somePath
    `sp'` \s -> someLocalDir `andDo` backupLocal s
  , someString "set" `sp` someString "rescan" `sp` someBool `andDo` ( lift . ( st_withRescan .= ) )
  , someString "set" `sp` someString "verbosity" `sp` someVerbosity `andDo` ( lift . ( st_verbosity . verb_level .= ) )
  , someString "debug" `andDo` const debug
  , someString "label" `sp` someString "--environment" `sp` someEnvid `sp'` \ env -> someLabel >>= \ label -> return ( lift $ labelEnv env label )
  , someString "label" `sp` someString "--storage" `sp` someSid `sp'` \ sid -> someLabel >>= \ label -> return ( lift $ labelStorage sid label )
  ]
  where
    isLocal Local = True
    isLocal _ = False
    inLocal :: Monad m => IParser u ( Interaction m ) a -> IParser u ( Interaction m ) a
    inLocal = condM $ use $ is_context . to isLocal
    inStorage = condM' $ use $ is_context . to icToSid
    scan' mpri = several someLocalPath `andDo` \ paths -> do
        cwd <- use is_local_path
        toAdd <- mapM ( joinDirCanonical cwd ) paths
        lift $ addFiles mpri toAdd
    csHelp = someStringHelp "cs" "change current storage (no arg for local, '-' for previous)"
    ctHelp = someStringHelp "ct" "change reference time (e.g. 2016-12-12)"
    backupHelp = someStringHelp "backup" "save some files to backup storage"
    restoreHelp = someStringHelp "restore" "restore to local directory"
    symlinkHelp = someStringHelp "symlink" "restore to local directory with symlinks"
    backupStorage sid source dest = do
      rsource <- use is_path >>= \ d -> return $ dropWhile ( == '/' ) ( joinDir d source )
      destpath <- use is_local_path >>= lift . flip joinDirCanonical dest
      lift $ backupPath sid rsource destpath
    backupLocal source dest = do
      rsource <- use is_path >>= \ d -> return $ joinDir d source
      mrpath <- lift $ getStorage rsource
      destpath <- use is_local_path >>= lift . flip joinDirCanonical dest      
      case mrpath of
        Just ( RichPath sid _ rpath _ ) -> do
          outputMsg $ "Warning: backup from " ++ sid ++ ":" ++ rpath ++ " as of last scan"
          outputMsg $ "To backup current contents, perform a scan first"
          lift $ backupPath sid rpath destpath
        Nothing ->
          outputMsg "The path is not contained in a known storage. Scan it first"


runShell :: (MonadException m, DelFSMonad m, SymlinkFSMonad m,  BufferMonad m)
            => EnvironmentId -> TopF4LMonad InteractionKnowledge m ()
runShell envid = do
  init <- initialInteractionStatus
  externalLoop Nothing init
    where
      externalLoop mhist st = do
        ( newHist, newSt ) <- inSession envid $ flip runStateT st $ runInputT ( setComplete iComplete defaultSettings ) $ do
          case mhist of
            Just hist -> putHistory hist
            _ -> return ()
          internalLoop
          getHistory
        case ( newSt ^. is_commit, newSt ^. is_quit ) of
          ( True, False ) -> externalLoop ( Just newHist ) ( is_quit .~ False $ newSt )
          ( _, True ) -> return ()
          ( False, False ) -> error "Abnormal exit"
      internalLoop = withInterrupt $ do
        msid <- lift $ use is_sid
        path <- lift $ use is_path
        ts <- lift $ use is_time
        let timeLabel = case ts of
              At t | t == endOfTime -> ""
              At t -> '@' : timestampToLocaltime t
              Merged -> "*"
        let prompt = maybe "" (\ x -> "[" ++ abbrSid x ++ timeLabel ++ "] ") msid ++ path ++ " > "
        minput <- handleInterrupt ( outputStrLn "Ctrl-C pressed. Type quit to exit" >> return Nothing ) $ getInputLine prompt

        case minput of
          Nothing ->
            lift $ is_quit .= True
          Just "" ->
            return ()
          Just line -> do
            parse <- lift $ runParserT (fullParser parseCmd) () "" line
            case parse of
              Left err ->
                outputStrLn $ "Unable to parse command " ++ show err
              Right act ->
                lift $ do
                  k <- lift $ use st_knowledge
                  return k
                  act
        toQuit <- lift $ use is_quit
        toCommit <- lift $ use is_commit
        unless ( toQuit || toCommit ) internalLoop
