{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE FlexibleContexts #-}
module F4L.Interactive.IParser where

import F4L
import F4L.F4LMonad( VerbosityLevel(..) )
import F4L.Utils( readN, localtimeToTimestamp )

import Control.Monad.State.Strict
import Data.Char( isSpace )
import Data.List( dropWhileEnd, isPrefixOf )
import Text.Parsec


-- | Type of completion to perform when parsing input as an incomplete command
data Complete = CompletePath
              | CompleteLocalPath
              | CompleteDir
              | CompleteLocalDir
              | CompleteSid
              | CompleteFid
              | CompleteEnvid
              | CompleteFixed [(String, Maybe String)] -- ^ ( completion, help )
              | CompleteTimeStamp
              deriving (Eq, Show, Ord)


(<+>) :: Stream s m t => ParsecT s u m [a] -> ParsecT s u m [a] -> ParsecT s u m [a]
a <+> b = do
  ares <- try ( lookAhead a ) <|> return []
  bres <- try ( lookAhead b ) <|> return []
  return $ ares ++ bres


getPos :: (Monad m, Stream String m Char) => ParsecT String u m Int
getPos = getParserState >>= return . (\ x -> x - 1) . sourceColumn . statePos


-- | Combination of a partial and a full parser
--   This is used to define in one go how to interpret a command
--   and what tab-completion to perform
data IParser u m a = IParser {
    partialParser :: ParsecT String u m [(Int, Complete)]
  , fullParser :: ParsecT String u m a
  }


terminated :: Monad m => IParser u m a -> IParser u m a
terminated ( IParser a b ) = IParser a ( b <<- many ( char ' ' ) <<- eof )


instance (Monad m) => Monad ( IParser u m ) where
  IParser a1 a2 >>= f =
    IParser
      ( try a1 <|> ( a2 >>= partialParser . f ) )
      ( a2 >>= fullParser . f )
  return a = IParser ( return [] ) ( return a )
  fail a = IParser ( fail a ) ( fail a )


instance MonadTrans ( IParser u ) where
  lift a = IParser ( return [] ) ( lift a )


instance (Monad m) => Applicative ( IParser u m ) where
  pure = return
  (<*>) = ap


instance (Monad m) => Functor ( IParser u m ) where
  fmap f xs  =  xs >>= return . f


someString a = someStringHelp' a Nothing


someStringHelp :: Monad m => String -> String -> IParser u m String
someStringHelp a b = someStringHelp' a ( Just b )


toEscape = " \\\t\n"

escapedChar :: ( Monad m, Stream String m Char ) => ParsecT String u m Char
escapedChar = try ( char '\\' >> anyChar ) <|> noneOf toEscape


isEscapedPrefix :: String -> String -> Maybe String
isEscapedPrefix pfx s = if pfx `isPrefixOf` es then Just es else Nothing
  where
    es = concat $ [ if x `elem` toEscape then ['\\', x] else [x] | x <- s ]


unescape :: String -> String
unescape ( '\\' : x : xs ) = x : unescape xs
unescape ( '\\' : [] ) = []  -- TODO: raise error? return Nothing? return "\\"?
unescape ( x : xs ) = x : unescape xs
unescape [] = []


someStringHelp' :: Monad m => String -> Maybe String -> IParser u m String
someStringHelp' s help = IParser p f
  where
    p = do
      pos <- getPos
      many escapedChar
      eof
      return [(pos, CompleteFixed [(s, help)])]
    f = string s


somePath_ :: Monad m => Complete -> IParser u m String
somePath_ c = IParser p f
  where
    p = do
      pos <- getPos
      many escapedChar
      optional $ char '\\'
      eof
      return [(pos, c)]
    f = many1 escapedChar


somePath :: Monad m => IParser u m String
somePath = somePath_ CompletePath


someLocalPath :: Monad m => IParser u m String
someLocalPath = somePath_ CompleteLocalPath


someDir :: Monad m => IParser u m String
someDir = somePath_ CompleteDir


someLocalDir :: Monad m => IParser u m String
someLocalDir = somePath_ CompleteLocalDir


someNoSpace :: Monad m => Complete -> IParser u m String
someNoSpace complete = IParser p f
  where
    p = do
      pos <- getPos
      many $ noneOf " "
      eof
      return [(pos, complete)]
    f = many1 $ noneOf " "


someSid :: Monad m => IParser u m String
someSid = someNoSpace CompleteSid


someFid :: Monad m => IParser u m String
someFid = someNoSpace CompleteFid


someEnvid :: Monad m => IParser u m String
someEnvid = someNoSpace CompleteEnvid


someSize :: Monad m => IParser u m Int
someSize = IParser ( fail "No parse" ) $ do
  i <- many1 ( oneOf ['0' .. '9'] )
  s <- liftM ( : [] ) ( try ( oneOf "kKmMgG" ) ) <|> return ""
  return $ readN $ i ++ s


someTimeStamp :: Monad m => IParser u m TimeStamp
someTimeStamp = do
  lt <- someNoSpace CompleteTimeStamp
  case localtimeToTimestamp lt of
    Nothing -> if lt == "latest" then return endOfTime else fail ( "Invalid timestamp " ++ lt )
    Just ts -> return ts


someVerbosity :: Monad m => IParser u m VerbosityLevel
someVerbosity = foldl1 orElse [ someString ( show i ) >> return ( toEnum i ) | i <- [0 .. 3] ]


somePriority :: Monad m => IParser u m Priority
somePriority = 
  foldl1 orElse $ [
      someStringHelp "none" "don't request backups" >> return Skip
    , someStringHelp "auto" "standard" >> return DefaultPriority
    , someStringHelp "top" "highest priority" >> return AsManyAsPossible
    ] ++ [ someString ( show i ) >> return ( Priority i ) | i <- [1..10] ]


someBool :: Monad m => IParser u m Bool
someBool = foldl1 orElse [
    someString "true" >> return True
  , someString "false" >> return False
  ]


someLabel :: Monad m => IParser u m String
someLabel = IParser p f
  where
    p = return []
    f = liftM ( dropWhile isSpace . dropWhileEnd isSpace ) $ liftM2 (:) ( noneOf " " ) ( many1 anyChar )


condM' :: Monad m => m ( Maybe c ) -> ( c -> IParser u m a ) -> IParser u m a
condM' cond f = IParser ( cond' ( partialParser . f ) ) ( cond' ( fullParser . f ) )
  where
    cond' g = do
      c <- lift cond
      case c of
        Nothing -> fail ""
        Just v -> g v


condM a f = condM' b ( const f )
  where
    b = do
      v <- a
      return $ if v then Just () else Nothing


IParser a1 a2 `orElse` IParser b1 b2 = IParser (a1 <+> b1) (try a2 <|> b2)


IParser a1 a2 `andDo` f2 = IParser a1 (liftM f2 a2)


a <<- b = do
  res <- a
  b
  return res


sp' ( IParser a1 a2 ) f =
  IParser
    ( try a1 <|> ( ( a2 <<- many1 (char ' ') ) >>= partialParser . f ) )
    ( ( a2 <<- many1 (char ' ') ) >>= fullParser . f )


sp a = sp' a . const


several :: Monad m => IParser u m String -> IParser u m [ String ]
several a = ( a `sp'` \ x -> liftM ( x : ) $ several a ) `orElse` liftM ( : [] ) a
