{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE FlexibleInstances, RankNTypes, DeriveGeneric, DeriveAnyClass #-}

module F4L.Knowledge where

import F4L
import F4L.Op
import F4L.PathAttributes as PA
import F4L.FileTree as FT
import F4L.StoryTree as ST
import F4L.Utils

import Control.DeepSeq
import Lens.Micro.Platform
import Control.Monad
import Control.Monad.State.Strict
import qualified Data.Binary as B
import qualified Data.IntMap.Strict as IM
import qualified Data.Map.Strict as SM
import Data.Maybe( fromMaybe, mapMaybe )
import Data.Set( Set )
import qualified Data.Set as S
import qualified Data.Store as DS
import GHC.Generics( Generic )
import GHC.DataSize
import System.FilePath( joinPath, splitDirectories )


-- | Stricter version of at for maps
at' :: Ord k => k -> Lens' ( SM.Map k v ) ( Maybe v )
at' k = lens ( SM.lookup k ) ( \ m v -> SM.alter ( const v ) k m )


at'' :: Int -> Lens' ( IM.IntMap v ) ( Maybe v )
at'' k = lens ( IM.lookup k ) ( \ m v -> IM.alter ( const v ) k m )


data FileData = FileData { _fd_seen :: !( Set ( StorageId, ShortPath ) )
                         , _fd_priorities :: !( SM.Map Priority Int )
                         , _fd_attrs :: ![FileAttr]
                         }
                deriving ( Show, Generic, DS.Store )


fd_seen :: Lens' FileData ( Set ( StorageId, ShortPath ) )
fd_seen = lens _fd_seen $ \ s a -> s { _fd_seen = a }
{-# INLINE fd_seen #-}

fd_priorities :: Lens' FileData ( SM.Map Priority Int )
fd_priorities = lens _fd_priorities $ \ s a -> s { _fd_priorities = a }
{-# INLINE fd_priorities #-}

-- fd_priority :: Getter FileData Priority
fd_priority :: Getting a FileData Priority
fd_priority = fd_priorities . to f
  where
    f fd = case map fst $ filter ( ( > 0 ) . snd ) $ SM.toList $ fd of
      [] -> error $ "File with no assigned priority: " ++ show fd
      xs -> maximum xs
{-# INLINE fd_priority #-}

fd_withPriority :: Priority -> FileData -> FileData
fd_withPriority pri = fd_priorities %~ SM.insertWith (+) pri 1


fd_swapPriority :: Priority -> Priority -> FileData -> FileData
fd_swapPriority pre post fd =
  fd_withPriority post $ fd_priorities %~ SM.alter dec pre $ fd
  where
    dec Nothing = error $ "Trying to replace unset priority " ++ show pre ++ " for " ++ show fd
    dec ( Just i ) | i <= 0 = error $ "Wrong multiplicity for priority " ++ show pre ++ " in " ++ show fd 
    dec ( Just 1 ) = Nothing
    dec ( Just i ) = Just $! i - 1


fd_attrs :: Lens' FileData [FileAttr]
fd_attrs = lens _fd_attrs $ \ s a -> s { _fd_attrs = a }
{-# INLINE fd_attrs #-}


noFileData = FileData S.empty SM.empty []


type StorageSeenMap = SM.Map StorageId [ ( EnvironmentId, FilePath ) ]


storagesSeenIn envid = SM.map ( map snd . filter ( ( envid == ) . fst ) )


instance B.Binary FileData where
  put ( FileData a b c ) = B.put a >> B.put b >> B.put c
  get = liftM3 FileData B.get B.get B.get


instance NFData FileData where
  rnf ( FileData a b c ) = rnf a `seq` rnf b `seq` rnf c


-- | Information updated by facts
class Knowledge k where
  noKnowledge :: k
  applyFact :: FidStore -> Fact -> k -> k
  applyFact = const $ const id
  kCode :: k -> String
  sizeInfo :: k -> IO [ ( String, Word ) ]
  sizeInfo k = recursiveSize k >>= \ w -> return [ ( kCode k, w ) ]


instance Knowledge () where
  noKnowledge = ()
  kCode = const ""


instance ( Knowledge a, Knowledge b ) => Knowledge ( a, b ) where
  noKnowledge = ( noKnowledge, noKnowledge )
  applyFact fids f ( a, b ) = -- TODO: more concise way to force strictness?
    let 
      na = applyFact fids f a
      nb = applyFact fids f b
    in
     na `seq` nb `seq` ( na, nb )
  kCode ( a, b ) = "C" ++ kCode a ++ "_" ++ kCode b ++ "D"
  sizeInfo k @ ( a, b ) = do
    s_a <- sizeInfo a
    s_b <- sizeInfo b
    s_tot <- recursiveSize k >>= \ w -> return [ ( kCode k, w ) ]
    return $ s_a ++ s_b ++ s_tot


-- | A version of binary which can rely on an existing FidStore for better efficiency
class ( Knowledge k, B.Binary k, DS.Store k ) => SerializableKnowledge k where
  putK :: FidStore -> k -> B.Put
  putK = const B.put

  getK :: FidStore -> B.Get k
  getK = const B.get


instance ( SerializableKnowledge k1, SerializableKnowledge k2 ) => SerializableKnowledge (k1, k2) where
  putK f (k1, k2) = putK f k1 >> putK f k2
  getK f = do
    k1 <- getK f
    k2 <- getK f
    k1 `seq` k2 `seq` return $! (k1, k2)


---------------------------------------- KnownData

-- | Standard information on files and storages
data KnownData = KnownData { 
    _kd_files :: !(IM.IntMap FileData)
  , _kd_storages :: !StorageSeenMap
  , _kd_contents :: !(SM.Map StorageId (StoryTree FidIndex))
  , _kd_priorities :: !(SM.Map StorageId (PathNode (Maybe Priority)))
  } deriving ( Show, Generic, DS.Store )


-- | Types containing a KnownData-like substructure
class KnownDataClass a where
  kd_files :: Lens' a (IM.IntMap FileData)
  kd_storages :: Lens' a StorageSeenMap
  kd_contents :: Lens' a (SM.Map StorageId (StoryTree FidIndex))
  kd_priorities :: Lens' a (SM.Map StorageId (PathNode (Maybe Priority)))


instance KnownDataClass KnownData where
  kd_files = lens _kd_files $ \ s a -> s { _kd_files = a }
  {-# INLINE kd_files #-}
  kd_storages = lens _kd_storages $ \ s a -> s { _kd_storages = a }
  {-# INLINE kd_storages #-}
  kd_contents = lens _kd_contents $ \ s a -> s { _kd_contents = a }
  {-# INLINE kd_contents #-}
  kd_priorities = lens _kd_priorities $ \ s a -> s { _kd_priorities = a }
  {-# INLINE kd_priorities #-}


instance {-# OVERLAPPING #-} KnownDataClass (KnownData, b) where
  kd_files = _1 . kd_files
  {-# INLINE kd_files #-}
  kd_storages = _1 . kd_storages
  {-# INLINE kd_storages #-}
  kd_contents = _1 . kd_contents
  {-# INLINE kd_contents #-}
  kd_priorities = _1 . kd_priorities
  {-# INLINE kd_priorities #-}


instance {-# OVERLAPPING #-} KnownDataClass a => KnownDataClass (b, a) where
  kd_files = _2 . kd_files
  {-# INLINE kd_files #-}
  kd_storages = _2 . kd_storages
  {-# INLINE kd_storages #-}
  kd_contents = _2 . kd_contents
  {-# INLINE kd_contents #-}
  kd_priorities = _2 . kd_priorities
  {-# INLINE kd_priorities #-}


instance Knowledge KnownData where
  noKnowledge = KnownData IM.empty SM.empty SM.empty SM.empty
  applyFact = applyFact''
  kCode = const "k3"
  sizeInfo k @ ( KnownData a b c d ) = do
    let c a = kCode k ++ a
    s_a <- recursiveSize a
    s_b <- recursiveSize b
    s_c <- recursiveSize c
    s_d <- recursiveSize d
    s_tot <- recursiveSize k
    return [
        ( c ".files", s_a )
      , ( c ".storages", s_b )
      , ( c ".contents", s_c )
      , ( c ".priorities", s_d )
      , ( c "", s_tot )
      ]


instance SerializableKnowledge KnownData where
  putK _ ( KnownData a b c d ) = do
    B.put b
    B.put ( IM.toList a )
    B.put c
    B.put d
  getK _ = do
    b <- getMap getStrict getStrict
    let sids = SM.fromList [ (sid, sid) | sid <- SM.keys b ]
    let getSid sid = fromJust' "Deduplicate sid" $ SM.lookup sid sids
    let dedup = fd_seen %~ S.fromList . map ( mapFst getSid ) . S.toList
    a <- liftM IM.fromList $ getList $ do
      ( a, b ) <- B.get
      return $!! ( a, dedup b )
    c <- getMap getStrict getStrict
    d <- getMap getStrict getStrict
    return $! KnownData a b c d


applyFact'' :: FidStore -> Fact -> KnownData -> KnownData
applyFact'' fids ( FileSeen ts fid sid path ) = execState $ do
  pri <- use $ kd_priorities . at' sid . to ( fromMaybe PA.empty ) . pathLens True path . to ( fromMaybe DefaultPriority )
  -- TODO: optimize if already seen in same position
  let idx = fidIndex' fids fid
  kd_contents %= SM.insertWith ( const $ insertFile ts path idx ) sid ( insertFile ts path idx ST.empty )
  kd_files . at'' idx %= Just . fd_withPriority pri . addToSeen  . fromMaybe noFileData
    where
      addToSeen = fd_seen %~ S.insert ( sid, pathToShort path )
applyFact'' _fids ( FileMiss ts sid path ) = execState $ do
  mIdx <- use $ kd_contents . at' sid . to ( >>= ST.findFileNodeAtTime ts path )
  kd_contents . at' sid %= liftM ( removeFile ts path )
  case mIdx of
    Just idx -> kd_files . at'' idx . traverse . fd_seen %= S.delete (sid, pathToShort path)
    Nothing -> return ()
applyFact'' fids ( FileAttr _ fid attrs ) =
  let idx = fidIndex' fids fid
  in
    execState $ kd_files . at'' idx . traverse . fd_attrs %= \ old -> old `seq` ( attrs ++ old )
applyFact'' _fids ( PathAttr _ sid path attrs ) =
  case mapMaybe ( \ a -> case a of Require pri -> Just pri; _ -> Nothing ) attrs of
    [] -> id
    pri : _ -> execState $ do
      oldPris <- use $ kd_priorities . at' sid . to ( fromMaybe PA.empty )
      -- TODO: don't use atTime
      subTree <- use $ kd_contents . at' sid . to ( fromMaybe ST.empty ) . to ( findNode ( splitDirectories path ) . atTime endOfTime ) . to ( fromMaybe FT.empty )
      let pathsFids = map ( mapFst $ joinPath . ( path : ) ) $ allFiles subTree
      forM_ pathsFids $ \ ( path, fid ) ->
        kd_files . at'' fid %= fmap ( fd_swapPriority ( fromMaybe DefaultPriority $ oldPris ^. pathLens True path ) pri )
      kd_priorities . at' sid .= Just ( pathSet path .~ PathNode ( Just pri ) SM.empty $ oldPris )
applyFact'' _fids ( StorageSeen _ts envid sid path ) = kd_storages %~ storageInsert envid sid path
  where
    storageInsert :: EnvironmentId -> StorageId -> FilePath -> StorageSeenMap -> StorageSeenMap
    storageInsert envid sid fp m =
      flip ( SM.insert sid ) m $ case SM.lookup sid m of
        Nothing -> [ ( envid, fp ) ]
        Just l -> ( envid, fp ) : filter ( ( envid, fp ) /= ) l
applyFact'' _fids _ = id


instance B.Binary KnownData where
  put ( KnownData a b c d ) = B.put a >> B.put b >> B.put c >> B.put d
  get = liftM4 KnownData B.get B.get B.get B.get


---------------------------------------- FactCounter
data FactCounter = FactCounter { fc_c :: !Int }
                  deriving (Show, Eq, Generic, DS.Store)


class FactCounterClass a where
  fact_counter :: Lens' a Int


instance {-# OVERLAPPING #-} FactCounterClass FactCounter where
  fact_counter = lens fc_c $ const FactCounter
  {-# INLINE fact_counter #-}


instance {-# OVERLAPPING #-} FactCounterClass ( FactCounter, b ) where
  fact_counter = _1 . fact_counter
  {-# INLINE fact_counter #-}


instance {-# OVERLAPPING #-} FactCounterClass a => FactCounterClass ( b, a ) where
  fact_counter = _2 . fact_counter
  {-# INLINE fact_counter #-}


instance Knowledge FactCounter where
  noKnowledge = FactCounter 0
  applyFact _ _ ( FactCounter i ) = FactCounter ( i + 1 )
  kCode = const "c"


instance B.Binary FactCounter where
  put = B.put . fc_c
  get = liftM FactCounter B.get


instance SerializableKnowledge FactCounter


---------------------------------------- FileSet
data FileSet = FileSet {
    _fs_files :: !( S.Set FidIndex )
  , _fs_size :: !Int
  } deriving (Show, Generic, DS.Store)


class FileSetClass a where
  fs_files :: Lens' a ( S.Set FidIndex )
  fs_size :: Lens' a Int


instance FileSetClass FileSet where
  fs_files = lens _fs_files $ \ s a -> s { _fs_files = a }
  fs_size = lens _fs_size $ \ s a -> s { _fs_size = a }


instance FileSetClass ( FileSet, b ) where
  fs_files = _1 . fs_files
  fs_size = _1 . fs_size


instance FileSetClass a => FileSetClass ( b, a ) where
  fs_files = _2 . fs_files
  fs_size = _2 . fs_size


instance Knowledge FileSet where
  noKnowledge = FileSet S.empty 0
  applyFact fids f x @ ( FileSet fs s ) = case mfact_file f of
    Just fid -> do
      let idx = fidIndex' fids fid
      if S.member idx fs then x else
        FileSet ( S.insert idx fs ) ( s + fid_length fid )
    Nothing -> x
  kCode = const "f"


instance B.Binary FileSet where
  put ( FileSet fs s ) = B.put fs >> B.put s
  get = liftM2 FileSet B.get B.get


instance SerializableKnowledge FileSet


---------------------------------------- SeenStorages
data SeenStorages = SeenStorages {
  _ss_storages :: !( SM.Map StorageId [ String ] )
  } deriving (Show, Generic, DS.Store)


class SeenStoragesClass a where
  seen_storages :: Lens' a ( SM.Map StorageId [ String ] )


instance {-# OVERLAPPING #-} SeenStoragesClass SeenStorages where
  seen_storages = lens _ss_storages $ \ s a -> s { _ss_storages = a }


instance {-# OVERLAPPING #-} SeenStoragesClass ( SeenStorages, b ) where
  seen_storages = _1 . seen_storages


instance {-# OVERLAPPING #-} SeenStoragesClass a => SeenStoragesClass ( b, a ) where
  seen_storages = _2 . seen_storages


instance Knowledge SeenStorages where
  noKnowledge = SeenStorages SM.empty
  applyFact _ ( LabelStorage sid label ) ( SeenStorages s ) =
    SeenStorages ( SM.insertWith (++) sid [ label ] s )
  applyFact _ x ( SeenStorages s ) =
    case mfact_storage x of
      Just sid -> sid `deepseq` SeenStorages ( SM.insertWith (++) sid [] s )
      Nothing -> SeenStorages s
  kCode = const "s"


instance B.Binary SeenStorages where
  put = B.put . _ss_storages
  get = liftM SeenStorages B.get


instance SerializableKnowledge SeenStorages


---------------------------------------- SeenEnvironments
data SeenEnvironments = SeenEnvironments {
  _se_environments :: !( SM.Map StorageId [ String ] )
  } deriving (Show, Generic, DS.Store)


class SeenEnvironmentsClass a where
  seen_environments :: Lens' a ( SM.Map StorageId [ String ] )


instance SeenEnvironmentsClass SeenEnvironments where
  seen_environments = lens _se_environments $ \ s a -> s { _se_environments = a }


instance {-# OVERLAPPING #-} SeenEnvironmentsClass ( SeenEnvironments, b ) where
  seen_environments = _1 . seen_environments


instance SeenEnvironmentsClass a => SeenEnvironmentsClass ( b, a ) where
  seen_environments = _2 . seen_environments


instance Knowledge SeenEnvironments where
  noKnowledge = SeenEnvironments SM.empty
  applyFact _ ( LabelEnvironment envid label ) ( SeenEnvironments s ) =
    SeenEnvironments ( SM.insertWith (++) envid [ label ] s )
  applyFact _ x @ ( StorageSeen {} ) ( SeenEnvironments s ) =
    SeenEnvironments ( SM.insertWith (++) ( fact_env x ) [] s )
  applyFact _ _ s = s
  kCode = const "e"


instance B.Binary SeenEnvironments where
  put = B.put . _se_environments
  get = liftM SeenEnvironments B.get


instance SerializableKnowledge SeenEnvironments
