{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE FlexibleInstances, RankNTypes, DeriveGeneric, DeriveAnyClass #-}


module F4L.StoryTree (
    StoryTree
  , TimeSpec(..)
  , allTimes
  , atTime
  , atTimeSpec
  , empty
  , findFileNodeAtTime
  , findNodeAtTime
  , fromList
  , getStoryFileTree
  , insertFile
  , merged
  , setPathAtTime
  , removeFile
  , toTree
  ) where

import F4L
import qualified F4L.FileTree as FT
import F4L.Utils

import Control.DeepSeq
import Control.Monad
import Data.Binary
import Data.Function( on )
import Data.Function.Pointless( (.!) )
import Data.List( find, inits )
import Data.Maybe( mapMaybe, listToMaybe, maybeToList )
import Data.Map.Strict hiding ( filter, fromList, map, mapMaybe, null, empty )
import qualified Data.Map.Strict as SM
import qualified Data.Set as Set
import qualified Data.Store as DS
import Data.Tree( Tree(..) )
import GHC.Generics( Generic )
import System.FilePath( splitDirectories )


newtype Story a = Story { unstory :: [ (TimeStamp, a) ] }
  deriving ( Eq, Ord, Generic, DS.Store )

instance Show a => Show ( Story a ) where
  show ( Story xs ) = show xs


instance Functor Story where
  fmap f = Story . map ( mapSnd f ) . unstory


noStory :: Story a
noStory = Story []


atTimeMaybe :: TimeStamp -> Story a -> Maybe a
atTimeMaybe ts = fmap snd . find ( ( <= ts ) . fst ) . unstory


latest :: Story a -> Maybe a
latest = fmap snd . listToMaybe . unstory


data TimeSpec = At TimeStamp | Merged


atTimeSpec :: Eq a => TimeSpec -> StoryTree a -> FT.FileTree a
atTimeSpec ( At ts ) = atTime ts
atTimeSpec Merged = merged


stricthead :: [(a, b)] -> [(a, b)]
stricthead x @ ( (a, b) : _ ) = a `seq` b `seq` x
stricthead [] = []


setAtTime :: Eq a => TimeStamp -> a -> Story a -> Story a
setAtTime ts a ( Story s ) = Story $! stricthead $ f' s
  where
    f' [] = [(ts, a)]
    f' s @ ( (t0, x0) : _ ) | t0 <= ts && x0 == a = trace "replacing with itself" s
    f' ( (t0, _) : tss ) | t0 == ts = ( (t0, a) : tss )
    f' tss @ ( (t0, _) : _ ) | t0 < ts = (ts, a) : tss
    f' ( (t0, _) : _ ) | t0 > ts = error $ "Applying change with ts=" ++ show ts ++ " after t0=" ++ show t0


modifyAtTime :: a -> TimeStamp -> ( a -> a ) -> Story a -> Story a
modifyAtTime a ts f = Story .! stricthead . f' .! unstory
  where
    f' [] = [(ts, f a)]
    f' ( (t0, x0) : tss ) | t0 == ts = ( (t0, f x0) : tss )
    f' tss @ ( (t0, _) : _ ) | t0 < ts = (ts, f a) : tss
    f' ( (t0, _) : _ ) | t0 > ts = error $ "Applying change with ts=" ++ show ts ++ " after t0=" ++ show t0


-- | Filesystem-like tree assigning a value of type `a` to nodes at a certain time
data StoryTree a = Void
                | File !a
                | Dir !( SM.Map ShortPath ( Story ( StoryTree a ) ) )
                deriving (Eq, Ord, Show, Generic, DS.Store)


toTree :: Show a => StoryTree a -> Tree String
toTree ( Dir sm ) = Node "DIR" $ map toNode $ toList sm
  where
    toNode (path, s) = Node ( show path ) [ Node ( '@' : show ts ) [ toTree v ] | (ts, v) <- unstory s ]
toTree x = Node ( show x ) []


allTimes :: StoryTree a -> [ TimeStamp ]
allTimes = Set.toAscList . allTimes'


allTimes' :: StoryTree a -> Set.Set TimeStamp
allTimes' ( Dir sm ) = Set.unions $ map allTimes'' $ SM.elems sm
  where
    allTimes'' s = Set.unions $ Set.fromList( map fst ( unstory s ) ) : map ( allTimes' . snd ) ( unstory s )
allTimes' _ = Set.empty


instance Functor StoryTree where
  fmap f ( File a ) = File ( f a )
  fmap f ( Dir d ) = Dir $ SM.map ( fmap $ fmap f ) d
  fmap f Void = Void


instance Binary a => Binary ( Story a ) where
  put = put . unstory
  get = liftM Story get


instance NFData a => NFData ( Story a ) where
  rnf = rnf . unstory


instance ( Binary a, NFData a ) => Binary ( StoryTree a ) where
  put ( File a ) = a `seq` put 'F' >> put a
  put ( Dir a ) = a `seq` put 'D' >> put a
  put Void = put 'V'
  get = do
    c <- get
    case c of
      'F' -> liftM ( File $!! ) get
      'D' -> do
        m <- get
        rnf m `seq` return ( Dir m )
      'V' -> return Void


instance NFData a => NFData ( StoryTree a ) where
  rnf ( File a ) = rnf a
  rnf ( Dir sm ) = rnf sm
  rnf Void = ()


getStoryFileTree :: Binary a => ( a -> FileId ) -> Get ( StoryTree FileId )
getStoryFileTree f = do
  c <- get
  case c of
    'F' -> liftM ( \ a -> File $!! f a ) get
    'D' -> liftM ( Dir $!! ) $ getMap getStrict $ getStory $ getStoryFileTree f
    'V' -> return Void
  where
    getStory :: Get a -> Get ( Story a )
    getStory getEl = do
      n <- get :: Get Int
      liftM Story $ getMany ( liftM2 (,) get getEl ) [] n

    getMany getEl xs 0 = return $! reverse xs
    getMany getEl xs i = do
      x <- getEl
      x `seq` getMany getEl (x:xs) (i-1)


empty :: StoryTree a
empty = Void


emptyDir :: StoryTree a
emptyDir = Dir SM.empty


fromList :: [( FilePath, Story ( StoryTree a ) )] -> StoryTree a
fromList = Dir . SM.fromList . map ( mapFst pathToShort )


setPathAtTime :: Eq a => TimeStamp -> [ ShortPath ] -> StoryTree a -> StoryTree a -> StoryTree a
setPathAtTime _ [] a _ = a
setPathAtTime ts ( x : xs ) a ( Dir sm ) = Dir $ SM.alter g x sm
  where
    -- g :: Maybe ( Story StoryTree ) -> Maybe ( Story StoryTree )
    g Nothing = g $ Just noStory
    g ( Just s ) = Just $ modifyAtTime Void ts ( setPathAtTime ts xs a ) s
setPathAtTime ts ( x : xs ) a _ = Dir $ SM.fromList [ ( x, setAtTime ts ( setPathAtTime ts xs a Void ) noStory ) ]


insertFile :: TimeStamp -> FilePath -> a -> StoryTree a -> StoryTree a
insertFile ts fp fid = insertLeaf ts ( map pathToShort $ splitDirectories fp ) ( File fid )


removeFile :: TimeStamp -> FilePath -> StoryTree a -> StoryTree a
removeFile ts fp = insertLeaf ts ( map pathToShort $ splitDirectories fp ) Void


insertLeaf :: TimeStamp -> [ShortPath] -> StoryTree a -> StoryTree a -> StoryTree a
insertLeaf _ [] a _ = a
insertLeaf ts ( x : xs ) a ( Dir sm ) = Dir $! SM.alter g x sm
  where
    -- g :: Maybe ( Story StoryTree ) -> Maybe ( Story StoryTree )
    g Nothing = g $ Just noStory
    g ( Just s ) = case unstory s of
      [] -> Just $ modifyAtTime Void ts ( insertLeaf ts xs a ) s
      y @ ( (t0, y0 @ ( Dir _ )) : ys )
        | t0 > ts ->
          error $ "Applying change with ts=" ++ show ts ++ " after t0=" ++ show t0 
        | not ( null xs ) ->
          Just $ Story $ stricthead $ (t0, insertLeaf ts xs a y0) : ys
        | null xs ->
          Just $ Story $ stricthead $ (ts, a) : y
      y @ ( (t0, y0) : ys )
        | t0 > ts ->
          error $ "Applying change with ts=" ++ show ts ++ " after t0=" ++ show t0 
        | t0 == ts ->
          Just $ Story $ stricthead $ (ts, insertLeaf ts xs a emptyDir ) : ys
        | otherwise ->
          Just $ Story $ stricthead $ (ts, insertLeaf ts xs a emptyDir ) : y
insertLeaf ts ( x : xs ) a _ = insertLeaf ts ( x : xs ) a emptyDir


findNodeAtTime :: TimeStamp -> [FilePath] -> StoryTree a -> Maybe ( StoryTree a )
findNodeAtTime _ [] st = Just st
findNodeAtTime _ _ Void = Nothing
findNodeAtTime _ _ ( File _ ) = Nothing
findNodeAtTime ts ( x : xs ) ( Dir st ) = SM.lookup ( pathToShort x ) st >>= atTimeMaybe ts >>= findNodeAtTime ts xs


findFileNodeAtTime :: TimeStamp -> FilePath -> StoryTree a -> Maybe a
findFileNodeAtTime ts path st =
  case findNodeAtTime ts ( splitDirectories path ) st of
    Just ( File x ) -> Just x
    _ -> Nothing


atTime :: TimeStamp -> StoryTree a -> FT.FileTree a
atTime t ( Dir sm ) = FT.Dir $ SM.mapMaybeWithKey f sm
  where
    f path s = case atTimeMaybe t s of
      Nothing -> Nothing
      Just d @ ( Dir _ ) ->
        case atTime t d of
          FT.Dir d' | SM.null d' -> Nothing
          x -> Just x
      Just ( File f ) -> Just $ FT.File f
      Just Void -> Nothing
atTime _ _ = FT.empty


data Flattened a = Flattened ( SM.Map ShortPath ( Maybe ( Flattened a ), Story a ) )


flatten :: StoryTree a -> Flattened a
flatten ( Dir m ) = Flattened $ SM.map flatten' m
flatten _ = Flattened $ SM.empty


flatten' :: Story ( StoryTree a ) -> ( Maybe ( Flattened a ), Story a )
flatten' ( Story s ) = ( merge ( mapMaybe get_dir s ), Story $ mapMaybe get_file s )
  where
    merge :: [ SM.Map ShortPath ( Story ( StoryTree a ) ) ] -> Maybe ( Flattened a )
    merge [] = Nothing
    merge dirs = Just $ Flattened $ SM.map flatten' $ SM.unionsWith concatStories dirs
    concatStories ( Story a ) ( Story b ) = Story $ a ++ b
    get_dir (_, Dir d) = Just d
    get_dir _ = Nothing
    get_file (ts, File f) = Just (ts, f)
    get_file _ = Nothing


disambiguate :: Eq a => Flattened a -> FT.FileTree a
disambiguate ( Flattened m ) = FT.Dir $ SM.fromList $ concatMap f $ SM.toList m
  where
    f (path, ( mf, s )) =
      let
        dirPart = map ( \ d -> ( path, disambiguate d ) ) $ maybeToList mf
        pathCandidates :: [ [ ShortPath ] ]
        pathCandidates = case mf of
          Nothing -> [ path ] : repeat []
          _ -> repeat []
        tsFile pathCandidates ( ts, file ) = (
          head $ pathCandidates ++ filter ( not . flip SM.member m ) ( fileCandidates ts file )
          , FT.File file
          )
        fileCandidates ts file = map ( pathToShort . ( ( shortToPath path ++ "." ++ timestampToLocaltime ts ) ++ ) ) ( inits $ repeat '~' )
      in
        dirPart ++ map ( uncurry tsFile ) ( zip pathCandidates $ reverse $ sNubBy ( (==) `on` snd ) $ reverse $ unstory s )


-- | All history merged into a single FileTree
merged :: Eq a => StoryTree a -> FT.FileTree a
merged = disambiguate . flatten
