{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
{-# LANGUAGE FlexibleContexts #-}

module F4L.Test (
  runTest
  ) where

import F4L
import F4L.Actions
import F4L.BufferMonad
import F4L.FidStore as FS
import F4L.FileTree as FT
import F4L.FSMonad hiding( Error )
import F4L.Knowledge
import F4L.Interactive
import F4L.MockFSMonad
import F4L.Op
import F4L.StoryTree as ST hiding( removeFile )
import F4L.Utils

import Control.DeepSeq( deepseq, force, rnf )
import Control.Monad.Trans.Except
import Control.Monad.State.Strict
import Lens.Micro.Platform
import qualified Data.Binary as B
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BL
import qualified Data.IntMap.Strict as IM
import Data.List( foldl', inits, intercalate, sort, (\\) )
import Data.Maybe( catMaybes, fromMaybe, fromJust )
import qualified Data.Map.Strict as SM
import Data.Set( Set )
import qualified Data.Set as S
import System.Console.Haskeline( replacement )
import System.IO( hPutStr, stderr )


fid i = FileId i $! force $ sha1 $ s2bl $ show i


s2bs s = BS.pack $ map ( toEnum . fromEnum ) s
s2bl s = BL.pack $ map ( toEnum . fromEnum ) s


data TTest a = Fast { shortcut :: String, ttest :: a } | Slow { shortcut :: String, ttest :: a }


isFast ( Fast {} ) = True
isFast _ = False


tests :: [TTest (Test String)]
tests = [
    Fast "fi" testFidInsert
  , Fast "fc" testFidConcat
  , Fast "jd" testJoinDir
  , Fast "af100" $ testAddFiles 100
  , Slow "af5e4" $ testAddFiles 50000
  , Fast "a1f" $ testAdd1File
  , Fast "af2s100" $ testAddFiles2Step 100
  , Slow "af2s5e4" $ testAddFiles2Step 50000
  , Fast "arf1e3" $ testAddRemoveFiles 1000
  , Slow "arf1e4" $ testAddRemoveFiles 10000
  , Slow "arf1e5" $ testAddRemoveFiles 100000
  , Fast "ms10x10" $ testMultiScan 10 10
  , Fast "ms10x100" $ testMultiScan 10 100
  , Slow "ms100x1000" $ testMultiScan 100 1000
  , Slow "ms100x10000" $ testMultiScan 100 10000
  , Fast "res" testRestore
  , Fast "mres-100-10" $ testMultiRestore 100 10
  , Slow "mres-1e4-10" $ testMultiRestore 10000 10
  , Slow "mres-1e5-100" $ testMultiRestore 100000 100
  , Fast "rescan" testWithRescan
  , Fast "re1" testRequirements
  , Fast "kd-1e4" $ testKnownData 10000
  , Slow "kd-1e5" $ testKnownData 100000
  , Slow "kd-3e5" $ testKnownData 300000
  , Slow "kd-1e6" $ testKnownData 1000000
  , Fast "c-cwd" testCwdCompletion
  , Fast "c-fn" testFnCompletion
  , Fast "c-st-100" $ testStorageCompletion 100
  , Slow "c-st-1e4" $ testStorageCompletion 10000
  , Fast "st" $ testStoryTree 100
  , Fast "stov-100" $ testStoryTreeOverwriting 100
  , Fast "stov-1e4" $ testStoryTreeOverwriting 10000
  , Fast "st-100" $ testStoryTreeGeneration 100
  , Fast "st-1e4" $ testStoryTreeGeneration 10000
  , Slow "st-1e5" $ testStoryTreeGeneration 100000
  , Slow "st-1e6" $ testStoryTreeGeneration 1000000
  , Fast "st-m" testMergedStoryTree
  , Fast "k-ser-b" $ testKnowledgeSerialization BinaryCache
  , Fast "k-ser-s" $ testKnowledgeSerialization StoreCache
  , Fast "cache-b" $ testCache BinaryCache
  , Fast "cache-s" $ testCache StoreCache
  ]


runTest :: [String] -> IO ()
runTest ["list"] = mapM_ putStrLn [ ( reverse $ take 2 $ reverse $ ' ' : show i ) ++ " - " ++ shortcut t ++ " - " ++ test_name ( ttest t ) | (i, t) <- zip [1..] tests ]
runTest ["all"] = runSomeTests tests
runTest [] = runSomeTests $ filter isFast tests
runTest xs = runSomeTests [ fromMaybe ( error $ x ++ " not found" ) $ lookup x els | x <- xs ]
  where
    els =  join [[(shortcut t, t), (show i, t)] | (i, t) <- zip [1..] tests ]


runSomeTests someTests = do
  res <- mapM ( run_it . pair shortcut ttest ) someTests
  hPutStr stderr $ "\r\ESC[KPassed " ++ show ( sum $ map snd res ) ++ "/" ++ show ( length res ) ++ " tests\n"
  case map fst $ filter ( (== 0) . snd ) res of
    [] -> return ()
    failed -> putStrLn $ "\nFailed: " ++ unwords failed
  where
    pair f g a = ( f a, g a )
    run_it ( sc, Test label act ) = do
      hPutStr stderr $ "\r\ESC[KRunning test: " ++ label
      res <- runExceptT act
      case res of
        Right _ -> hPutStr stderr " OK" >> return ( sc, 1 )
        Left msg -> hPutStr stderr ( "\nERR: " ++ msg ++ "\n" ) >> return ( sc, 0 )


data Test err = Test { test_name :: String, test_run :: ExceptT err IO () }


terminates :: Show a => String -> IO a -> Test err
terminates label act = Test label $ do
  a <- lift act
  return $ rnf ( show a )


defaultState :: F4LState KnownData ()
defaultState = f4lState (MetaDB "/tmp/no-f4l-db") (Verbosity Chatty True) UseDefault False False ()


testFidInsert = terminates "10000 fid insertions" $ do
  execStateT ( mapM step [1..10000] ) defaultState
  where
    step i = do
      let f = fid i
      show f `deepseq` do
        st_opEnv . openv_fid .= Just f
        st_opEnv . openv_fids %= withFid f


testFidConcat = terminates "10000 fid show" $ do
  return $ concatMap ( show . fid ) [ 1 .. 10000 ]


testJoinDir = Test "Join dir" $ mapM_ check [
  ( "/a/b", "..", "/a" )
  , ( "a/b", "..", "a" )
  , ( "/", "..", "/" )
  ]
  where
    check (a, b, c) =
      let j = joinDir a b
      in
       unless ( j == c ) ( throwE $ a ++ " join " ++ b ++ " = " ++ j ++ " <> " ++ c )


readFromFile' = onErrorWarnWithDefault "" . readFromFile


runMock :: ( MockFSState mfs, Knowledge k )
           => Int -> ExceptT err ( TopF4LMonad k ( MockFS mfs IO ) ) a -> ExceptT err ( StateT mfs IO ) a
runMock delta_t act =
  let
    v = Verbosity Quiet True
    state = f4lState (MetaDB "/home/user/.f4l/db") v UseDefault False False ()
  in ExceptT $ runMockFS $ evalStateT ( runExceptT $ ( lift $ lift $ mfs_now .= delta_t ) >> act ) state


inSession' :: ( FSMonad m, BufferMonad m, Knowledge k ) => F4LMonad k m a -> GenF4LMonad k s m a
inSession' = inSession "some_env"


vfs dir n = Overlay emptyMemFS $ VirtualFS dir n


vfs2 dir1 n1 dir2 n2 = Overlay emptyMemFS $ Overlay ( VirtualFS dir1 n1 ) ( VirtualFS dir2 n2 )


vfsFiles :: [ (FilePath, BL.ByteString) ] -> MemFS
vfsFiles = foldr ( uncurry mfs_set_contents ) emptyMemFS


mockTest :: MockFSState mfs
            => String -> mfs -> ExceptT err ( StateT mfs IO ) a -> Test err
mockTest label fs act = Test label $ void $ ExceptT $ evalStateT ( runExceptT act ) fs


data SeenFiles = SeenFiles {
    sf_seen :: !( Set FilePath ) }


instance Knowledge SeenFiles where
  noKnowledge = SeenFiles S.empty

  applyFact _ r @ ( FileSeen {} ) s =
    s { sf_seen = S.insert (fact_path r) $ sf_seen s }
  applyFact _ _ s = s

  kCode _ = "test-seenfiles"


interactionK :: InteractionKnowledge
interactionK = undefined


knownK :: KnownData
knownK = undefined


seenStoragesK :: SeenStorages
seenStoragesK = undefined


statsK :: ( SeenFiles, FactCounter )
statsK = undefined


testAddFiles n = mockTest ( "Adding " ++ hShow n ++ " files" ) ( vfs2 "/some/dir" n "/some/other" n ) ( runMock 0 act1 >> runMock 10 act2 )
  where
    act1 = do
      lift $ do
        withKnowledge interactionK
        st_interactive .= Use "2"
        inSession' $ addFiles Nothing ["/some/dir", "/some/other"]
      unlessM ( lift $ doesFileExist "/some/.f4l-id" ) $ throwE "No id file was created"
      unlessM ( lift $ doesFileExist "/home/user/.f4l/db" ) $ throwE "No id file was created"
    act2 = do
      lift $ withKnowledge statsK
      lift $ onErrorDie $ fromDB loadOps
      i <- lift $ use $ st_knowledge . fact_counter
      SeenFiles seen_paths <- lift $ use $ st_knowledge . _1
      forM_ ( liftM2 (++) ["dir/", "other/"] $ map show [1..n] ) $ \ f -> do
        unless ( f `S.member` seen_paths ) $ throwE $ f ++ " not found in facts"


testAddFiles2Step n = mockTest ( "Adding " ++ hShow n ++ " files - 2 step" ) ( vfs2 "/some/dir" n "/some/other" n ) ( runMock 0 act1 >> runMock 10 act2 >> runMock 20 act3 >> runMock 30 act4 )
  where
    act1 = do
      lift $ do
        withKnowledge interactionK
        st_interactive .= Use "2"
        inSession' $ addFiles Nothing ["/some/dir"]
      unlessM ( lift $ doesFileExist "/some/.f4l-id" ) $ throwE "No id file was created"
      unlessM ( lift $ doesFileExist "/home/user/.f4l/db" ) $ throwE "No id file was created"
    act2 = do
      lift $ onErrorDie $ fromDB loadOps
      FactCounter i <- lift $ use st_knowledge
      unless (i > n) $ throwE "Facts not found in DB"
    act3 = lift $ do
      withKnowledge interactionK
      inSession' $ addFiles Nothing ["/some/other"]
    act4 = do
      lift $ onErrorDie $ fromDB loadOps
      SeenFiles seen_paths <- lift $ use st_knowledge
      forM_ ( liftM2 (++) ["dir/", "other/"] $ map show [1..n] ) $ \ f -> do
        unless ( f `S.member` seen_paths ) $ throwE $ f ++ " not found in facts"


testAdd1File = mockTest ( "Adding 1 file" ) ( vfs2 "/some/dir" 10 "/some/dir/subdir" 10 ) ( runMock 0 act1 >> runMock 10 act2 )
  where
    act1 = do
      lift $ do
        withKnowledge interactionK
        st_interactive .= Use "2"
        inSession' $ addFiles Nothing ["/some/dir/1"]
      unlessM ( lift $ doesFileExist "/some/.f4l-id" ) $ throwE "No id file was created"
    act2 = do
      lift $ withKnowledge statsK
      lift $ onErrorDie $ fromDB loadOps
      i <- lift $ use $ st_knowledge . fact_counter
      SeenFiles seen_paths <- lift $ use $ st_knowledge . _1
      case S.size seen_paths of
        1 -> return ()
        0 -> throwE "No file was added. Expecting 1"
        _ -> throwE "Too many files were added. Expecting 1"


testAddRemoveFiles n = mockTest ( "Adding and removing " ++ hShow n ++ " files" ) ( vfs "/some/dir" n ) ( runMock 10 add >> runMock 20 del )
  where
    add = lift $ do
      withKnowledge interactionK
      inSession' $ addFiles Nothing [ "/some/dir" ]
    del = lift $ do
      withKnowledge interactionK
      sequence_ [
        onErrorWarn $ removeFile $ "/some/dir" ++ show i
        | i <- [ 1 .. n ]
        ]
      inSession' $ addFiles Nothing [ "/some/dir" ]


testMultiScan n m = mockTest ( "Adding " ++ hShow n ++ " files " ++ hShow m ++ " times" ) ( vfs "/some/dir" n ) ( sequence [ runMock t act | t <- [ 1 .. m ] ] )
  where
    act = do
      lift $ do
        withKnowledge interactionK
        st_interactive .= Use "2"
        inSession' $ addFiles Nothing ["/some/dir"]
      unlessM ( lift $ doesFileExist "/some/.f4l-id" ) $ throwE "No id file was created"
      unlessM ( lift $ doesFileExist "/home/user/.f4l/db" ) $ throwE "No id file was created"


testRestore = mockTest "Adding and restoring files" ( vfs "/some/dir" 100 ) ( runMock 0 act1 >> runMock 100 act2 >> runMock 200 act3 >> runMock 300 act4 )
  where
    act1 = lift $ do
      withKnowledge interactionK
      inSession' $ addFiles Nothing ["/some/dir"]
    act2 = do
      lift $ do
        withKnowledge interactionK
        sid <- readFromFile' "/some/dir/.f4l-id"
        onErrorDie $ fromDB loadOps
        inSession' $ onErrorWarn $ void $ restorePath True ( At endOfTime ) sid "" "/some/dir_copy"
      forM_ (map show [1..100]) $ \ f -> do
        orig <- lift $ readFromFile' $ "/some/dir/" ++ f
        copy <- lift $ readFromFile' $ "/some/dir_copy/" ++ f
        when ( orig /= copy ) $ throwE $ "File " ++ f ++ " not restored"
    act3 = lift $ do
      withKnowledge interactionK
      sequence_ [ onErrorWarn $ removeFile $ "/some/dir/" ++ show i | i <- [10, 20..100] ]
      onErrorDie $ fromDB loadOps
      inSession' $ addFiles Nothing ["/some/dir"]
    act4 = do
      lift $ do
        withKnowledge interactionK
        sid <- readFromFile' "/some/dir/.f4l-id"
        onErrorDie $ fromDB loadOps
        inSession' $ onErrorWarn $ void $ restorePath True ( At endOfTime ) sid "" "/some/dir_copy2"
      forM_ (map show $ [1..100] \\ [10, 20 .. 100]) $ \ f -> do
        orig <- lift $ readFromFile' $ "/some/dir/" ++ f
        copy <- lift $ readFromFile' $ "/some/dir_copy2/" ++ f
        when ( orig /= copy ) $ throwE $ "File " ++ f ++ " not restored"
      forM_ [ "/some/dir_copy2/" ++ show i | i <- [10, 20 .. 100] ] $ \ f -> do
        whenM ( lift $ doesFileExist f ) $ throwE $ "File " ++ f ++ " should not have been restored"


testMultiRestore n t = mockTest ( "Restoring " ++ hShow n ++ " files " ++ hShow t ++ " times" ) ( vfs "/some/dir" n ) ( runMock 0 act1 >>= \ sid -> forM_ [1..t] (\ i -> runMock (100 + i * 10) (act2 sid i)) )
  where
    act1 = lift $ do
      withKnowledge interactionK
      inSession' $ addFiles Nothing ["/some/dir"]
      readFromFile' "/some/dir/.f4l-id"
    act2 sid i = do
      lift $ do
        withKnowledge interactionK
        outputLineMsg $ "Restoring to " ++ show i
        inSession' $ onErrorWarn $ void $ restorePath True ( At endOfTime ) sid "" ( "/some/dir_" ++ show i )
        cleanLine


testWithRescan = mockTest "Reindexing files with rescan" ( vfs "/some/dir" 100 ) $ runMock 0 act1 >>= \ fid -> runMock 10 ( act2 fid ) >> runMock 20 ( act3 fid )
  where
    mutableFidx :: Monad m => GenF4LMonad InteractionKnowledge s m FidIndex
    mutableFidx = use $ st_knowledge . kd_contents . to ( head . SM.elems ) . to ( fromJust . FT.findFileNode "mutable" . atTime endOfTime )

    act1 = lift $ do
      withKnowledge interactionK
      onErrorWarn $ appendToFile "/some/dir/mutable" $ replicate 10000  'a'
      inSession' $ addFiles Nothing ["/some/dir"]
      mutableFidx
    act2 fid = do
      lift $ withKnowledge interactionK
      lift $ onErrorWarn $ do
        removeFile "/some/dir/mutable"
        appendToFile "/some/dir/mutable" $ replicate 10000 'A'
      lift $ do
        onErrorDie $ fromDB loadOps
        inSession' $ addFiles Nothing ["/some/dir"]
      n_fid <- lift mutableFidx
      when (n_fid /= fid) $ throwE "Expecting file change NOT to be detected"
    act3 fid = do
      lift $ do
        onErrorDie $ fromDB loadOps
        st_withRescan .= True
        inSession' $ addFiles Nothing ["/some/dir"]
      n_fid <- lift mutableFidx
      when (n_fid == fid) $ throwE "Expecting file change to be detected with rescan"


testRequirements = mockTest "Setting and changing requirements" ( vfs "/some/dir" 10 ) ( runMock 0 act1 >> runMock 10 act2 )
  where
    act1 = lift $ do
      withKnowledge interactionK
      inSession' $ addFiles Nothing ["/some/dir"]
    act2 = do
      lift $ withKnowledge interactionK
      f <- lift $ use $ st_knowledge . kd_files
      when ( any ( ( /= DefaultPriority ) . view fd_priority ) $ IM.elems f ) $ throwE "Added files don't have default priority"
      lift $ inSession' $ addFiles ( Just AsManyAsPossible ) ["/some/dir"]
      f <- lift $ use $ st_knowledge . kd_files
      when ( any ( ( /= AsManyAsPossible ) . view fd_priority ) $ IM.elems f ) $ throwE "Re-added files don't have max priority"


testKnownData n = terminates ( "Generating known data for " ++ hShow n ++ " files " ) act
  where
    act = hPutStr stderr $ show $ length $ show bigData
    boh :: KnownData
    boh = noKnowledge
    ini = ( FS.empty, boh )
    combine ( fs, k ) ( fid, fact ) =
      let newFs = withFid fid fs
      in ( newFs, applyFact newFs fact k )
    bigData = foldl' combine ini lotsOfFacts
    lotsOfFacts = flip map [1..n] $ \ i ->
      let fid = FileId i $ sha1 $ BL.pack $ map (toEnum . fromEnum) $ show i
      in ( fid, FileSeen 0 fid ( take 10 $ show i ++ "abcdefghi" ) ( "/some/path/" ++ show i ) )


testCwdCompletion = mockTest "Command completion on local dir" ( vfs "/some/dir" 100 ) ( runMock 0 act )
  where
    comp s = flip evalStateT s $ iComplete (reverse "ls /some/dir/", "")
    act = do
      iState <- lift initialInteractionStatus
      (_, completions) <- lift $ inSession' $ comp iState
      let rs = map replacement completions
      case filter (not . (`elem` rs)) ["/some/dir/" ++ show i | i <- [1..100]] of
        [] -> return ()
        x -> throwE $ show ( length x ) ++ "/100 files not present in completion"


testFnCompletion = mockTest "Filename completion" ( vfsFiles [( fn, BL.empty) | fn <- fns] ) ( runMock 0 act )
  where
    fns = ["/some/a b", "/some/f 1", "/some/f 2"]
    comp input s = flip evalStateT s $ iComplete (reverse ( "ls " ++ input ), "")
    checks = [
        ("/some/a\\ ", ["/some/a\\ b"])
      , ("/some/f", ["/some/f\\ 1", "/some/f\\ 2"])
      ]
    act' x @ (input, completion) = do
      iState <- lift initialInteractionStatus
      (_, completions) <- lift $ inSession' $ comp input iState
      let rs = map replacement completions
      return $ if ( sort rs /= sort completion ) then Just (x, rs) else Nothing
    act = do
      res <- mapM act' checks
      case catMaybes res of
        [] -> return ()
        withErr -> throwE $ "No completion for " ++ intercalate ", " ( map show withErr )


testStorageCompletion n = mockTest ( "Command completion on storage with " ++ hShow n ++ " files" ) ( vfs "/some/dir" n ) ( runMock 0 act1 >>= runMock 10 . act2 )
  where
    comp s = flip evalStateT s $ iComplete (reverse "ls /", "")
    act1 = do
      lift $ withKnowledge interactionK
      lift $ inSession' $ addFiles Nothing ["/some/dir"]
      esid <- lift $ readFromFile' "/some/dir/.f4l-id"
      if esid == "" then throwE "Sid not found in dir" else return esid
    act2 sid = do
      lift $ onErrorDie $ fromDB loadOps
      iState <- lift initialInteractionStatus
      (_, completions) <- lift $ inSession' $ comp $ set is_context ( InStorage sid ) iState
      let rs = map replacement completions
      case filter (not . (`elem` rs)) ["/" ++ show i | i <- [1..n]] of
        [] -> return ()
        x -> throwE $ show ( length x ) ++ "/" ++ show n ++ " files not present in completion"


assertEquals :: ( Eq a, Show a, Monad m ) => String -> a -> a -> ExceptT String m ()
assertEquals msg expected got =
  when ( expected /= got ) $ throwE $ "Mismatch in " ++ msg ++ ": expected " ++ show expected ++ ", got " ++ show got


testStoryTree :: Int -> Test String
testStoryTree n = Test "Change in StoryTree" $ do
  let st = foldl' ins ST.empty $ zip [1..] $ map testFid [1..n]
  unless ( ST.allTimes st == [1..n] ) $ throwE "Mismatch in allTimes"
  forM_ [1..n] $ \ ts -> assertEquals ( "number of files at time " ++ show ts ) ts ( length $ FT.allFiles $ atTime ts st )
  where
    ins st (i, a) = ST.insertFile i (take 1 ( show i ) ++ "/" ++ show i) a st


testStoryTreeOverwriting :: Int -> Test String
testStoryTreeOverwriting n = Test "Change in StoryTree" $ do
  let st = foldl' ins ST.empty $ zip [1..] $ map testFid [1..n]
  unless ( ST.allTimes st == [1..n] ) $ throwE "Mismatch in allTimes"
  forM_ [1..n] $ \ ts -> assertEquals ( "number of files at time " ++ show ts ) ( min ts 10 ) ( length $ FT.allFiles $ atTime ts st )
  assertEquals ( "number of merged files " ++ show n ) n ( length $ FT.allFiles $ merged st )
  where
    -- use i % 10 to compute path to force collisions
    ins st (i, a) = ST.insertFile i (take 1 ( show $ i `mod` 10 ) ++ "/" ++ show (i `mod` 10)) a st


testMergedStoryTree :: Test String
testMergedStoryTree = Test "Merging StoryTree" $ do
  let n = 10
  let st = foldl' ins ST.empty $ zip [1..] $ map testFid [1..n]
  assertEquals ( "Number of merged files = " ++ show n ) n ( length $ FT.allFiles $ merged st )
  where
    ins st (i, a) = ST.insertFile i (take 1 ( show i ) ++ "/" ++ show i) a st


testStoryTreeGeneration :: Int -> Test String
testStoryTreeGeneration n = terminates ( "Generating StoryTree with " ++ hShow n ++ " files" ) $ do
  let st = foldl' ins ST.empty $ zip [1..] $ map testFid [1..n]
  hPutStr stderr " computing length "
  hPutStr stderr $ hShow $ fromIntegral $ BL.length ( B.encode st )
  where
    ins st (i, a) = ST.insertFile i (intercalate "/" $ drop 1 $ inits $ show i) a st


testKnowledgeSerialization :: CacheSerialization -> Test String
testKnowledgeSerialization ser =
  mockTest ( "Knowledge serialization (" ++ show ser ++ ")" ) ( vfs "/some/dir" 1000 ) ( runMock 0 act1 )
  where
    act1 = do
      lift $ withKnowledge (seenStoragesK, knownK)
      lift $ inSession' ( addFiles Nothing ["/some/dir"] )
      hs <- liftM ( map $ \ (_, _, h) -> h ) $ fromDB $ return . sessionsWithHash
      lift $ onErrorDie $ saveCache ser hs
      lift $ onErrorDie $ loadCached ser 0


testCache :: CacheSerialization -> Test String
testCache ser =
  mockTest ( "Cache (" ++ show ser ++ ")" ) ( vfs "/some/dir" 0 ) ( runMock 0 act1 >> runMock 10 act2 >> runMock 20 act3 >> runMock 30 act4 )
  where
    act1 = lift $ do
      withKnowledge interactionK
      st_interactive .= Use "1"
      onErrorDie $ appendToFile "/some/dir/a" "a"
      inSession' $ addFiles Nothing ["/some/dir/a"]
    act2 = do
      -- /some/dir now contains `a` and `.f4l-id`
      unlessM ( lift $ doesFileExist "/some/dir/.f4l-id" ) $ throwE "No id file was created"
      lift $ do
        withKnowledge interactionK
        onErrorDie $ loadCached ser 0
        inSession' $ addFiles Nothing ["/some/dir"]
      fids <- lift $ use $ st_opEnv . openv_fids
      when ( fidByIndex fids 1 == Nothing ) $ throwE "Some FID was not inserted"
    act3 = do
      lift $ do
        withKnowledge interactionK
        onErrorDie $ loadCached ser 0
      fids <- lift $ use $ st_opEnv . openv_fids
      when ( fidByIndex fids 1 == Nothing ) $ throwE "Some FID was lost"
      lift $ do
        onErrorDie $ appendToFile "/some/dir/a2" "a"
        inSession' $ addFiles Nothing ["/some/dir"]
    act4 = do
      lift $ do
        withKnowledge interactionK
        onErrorDie $ loadCached ser 0
      fids <- lift $ use $ st_opEnv . openv_fids
      when ( fidByIndex fids 1 == Nothing ) $ throwE "Some FID was lost"
