{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014-2017
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L.Utils where

import qualified Codec.Binary.UTF8.String as UTF8
import Control.DeepSeq
import Control.Monad( liftM, when )
import Control.Monad.State.Strict
import qualified Crypto.Hash.SHA1 as SHA1
import Crypto.Hash.SHA1( hashlazy, Ctx, update, updates, finalize )
import qualified Data.Binary as B
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BL
import qualified Data.ByteString.Short as SBS
import qualified Data.IntMap as IM
import Data.List( find, isPrefixOf, sortBy, unfoldr )
import qualified Data.Map.Strict as SM
import Data.Map.Strict( Map )
import Data.Time.Clock.POSIX( posixSecondsToUTCTime, utcTimeToPOSIXSeconds )
import Data.Time.Format( formatTime, parseTime, defaultTimeLocale, iso8601DateFormat )
import Data.UUID( toString )
import Data.Word( Word8 )
import qualified System.Posix as P
import System.FilePath( splitDirectories )
import System.IO.Unsafe( unsafePerformIO )
import System.Random( randomIO )


trace :: Show a => a -> b -> b
trace msg v = unsafePerformIO $ print msg >> return v


traceit :: Show a => a -> a
traceit = \ a -> trace a a


hexChars :: IM.IntMap Char
hexChars = IM.fromList $ zip [0..] $ ['0' .. '9'] ++ ['a' .. 'f']


hashToS :: SBS.ShortByteString -> String
hashToS h = toHex $ SBS.unpack h
    where
      toHex = concatMap ( toHexB . fromEnum )
      toHexB a = toHexH ( a `div` 16 ) : toHexH ( a `mod` 16 ) : []
      {-# INLINE toHexB #-}
      toHexH :: Int -> Char
      toHexH b = fromJust' ( "Error in hex: got " ++ show b ) $ IM.lookup b hexChars
      {-# INLINE toHexH #-}


hashFromS :: String -> Maybe SBS.ShortByteString
hashFromS = liftM SBS.pack . sequence . unfoldr f
  where
    f :: String -> Maybe ( Maybe Word8, String )
    f ( a1 : a2 : rest ) = Just (
      do
           h1 <- readHex a1 
           h2 <- readHex a2
           return $ toEnum $ 16 * h1 + h2
      , rest )
    f [] = Nothing
    f _ = Just ( Nothing, [] )
    s = SM.fromList ( zip ( [ '0' .. '9' ] ++ [ 'a' .. 'f' ] ) [ 0 .. ] )
    readHex :: Char -> Maybe Int
    readHex c = SM.lookup c s


bl2bs :: BL.ByteString -> BS.ByteString
bl2bs = BS.concat . BL.toChunks


bs2bl :: BS.ByteString -> BL.ByteString
bs2bl = BL.pack . BS.unpack


sha1 :: BL.ByteString -> SBS.ShortByteString
sha1 dat = SBS.toShort $! hashlazy dat


newSha1 :: Ctx
newSha1 = SHA1.init


updateSha1 :: BL.ByteString -> Ctx -> Ctx
updateSha1 s c = updates c $ BL.toChunks s


updateSha1' :: BS.ByteString -> Ctx -> Ctx
updateSha1' s c = update c s


toSha1 :: Ctx -> SBS.ShortByteString
toSha1 c = SBS.toShort $ finalize c


readN :: String -> Int
readN s = case reverse s of
  [] -> error "No input for integer"
  p : revn ->
    case find ( ( p `elem` ) . fst ) [ ( "kK", 1000 ), ( "mM", 1000 * 1000 ), ( "gG", 1000 * 1000 * 1000 ) ] of
      Just ( _, mul ) -> round ( fromIntegral mul * read ( reverse revn ) )
      Nothing -> read s


subInsert :: ( Ord a, Ord b ) => a -> b -> v -> Map a ( Map b v ) -> Map a ( Map b v )
subInsert k1 k2 v m =
  -- k1 `seq` k2 `seq` v `seq` m `seq` 
  SM.insert k1 ( SM.insert k2 v $ SM.findWithDefault SM.empty k1 m ) m


fileSize :: Integral i => FilePath -> IO i
fileSize path =
  P.getFileStatus path >>= return . fromIntegral . P.fileSize


uuidString :: IO String
uuidString = randomIO >>= return . filter (/= '-') . toString


dominants :: ( a -> a -> Ordering ) -> ( a -> a -> Bool ) -> [a] -> [a]
dominants cmp dominates = sNubBy dominates . sortBy cmp


dominates (sid1, path1) (sid2, path2) =
  sid1 == sid2 && ( splitDirectories path1 `isPrefixOf` splitDirectories path2 )


hShow :: Int -> String
hShow = reverse . chunksOf 3 . reverse . show
  where
    chunksOf n xs = case splitAt n xs of
      (a, []) -> a
      (a, b) -> a ++ "," ++ chunksOf n b


whenM :: Monad m => m Bool -> m () -> m ()
whenM cond act = do
  c <- cond
  when c act


unlessM :: Monad m => m Bool -> m () -> m ()
unlessM = whenM . liftM not


leftToMaybe ( Left a ) = Just a
leftToMaybe _ = Nothing


-- | Remove consecutive duplicates
sNub :: Eq a => [a] -> [a]
sNub = sNubBy (==)


-- | Remove consecutive duplicates according to the given equality check
sNubBy :: ( a -> a -> Bool ) -> [a] -> [a]
sNubBy cmp (x : y : xs) | x `cmp` y = sNubBy cmp (x : xs)
sNubBy cmp (x : xs) = x : sNubBy cmp xs
sNubBy _ [] = []


mapFst f (a, b) = (f a, b)


mapSnd f (a, b) = (a, f b)


ifM :: Monad m => m Bool -> m a -> m a -> m a
ifM cond a b = do
  p <- cond
  if p then a else b


progressBar l i j = "[" ++ replicate t '=' ++ replicate ( l - t ) ' ' ++ "]"
  where t = l * i * 100 `div` j


ofType :: a -> a -> a
ofType = const


fromJust' :: String -> Maybe a -> a
fromJust' _ ( Just a ) = a
fromJust' msg _ = error $ "Nothing: " ++ msg


putList :: ( a -> B.Put ) -> [a] -> B.Put
putList putEl els = do
  B.put ( length els :: Int )
  mapM_ putEl els


getList :: B.Get a -> B.Get [a]
getList getOne = do
  n <- B.get :: B.Get Int
  go [] n
  where
    go xs 0 = return $! reverse xs
    go xs i = do
      x <- getOne
      x `seq` go (x:xs) (i-1)


getMap :: ( B.Binary a, B.Binary b, Ord a ) => B.Get a -> B.Get b -> B.Get ( Map a b )
getMap getKey getValue = liftM SM.fromList $ getList $ do
  k <- getKey
  v <- getValue
  k `seq` v `seq` return (k, v)


getMap1 :: ( B.Binary a, B.Binary b, NFData a, NFData b, Ord a ) => B.Get a -> B.Get b -> B.Get ( Map a b )
getMap1 getKey getValue = do
  n <- B.get :: B.Get Int
  g SM.empty n
  where
    g m 0 = return m
    g m i = do
      k <- getKey
      v <- getValue
      res <- g ( SM.insert k v m ) ( i - 1 )
      k `seq` v `seq` res `seq` return res


getStrict :: ( B.Binary a, NFData a ) => B.Get a
getStrict = liftM force B.get


type ShortPath = SBS.ShortByteString


pathToShort = SBS.pack . UTF8.encode


shortToPath = UTF8.decode . SBS.unpack


timestampToLocaltime :: Int -> String
timestampToLocaltime = formatTime defaultTimeLocale ( iso8601DateFormat $ Just "%H:%M:%S" ) . posixSecondsToUTCTime . fromIntegral


localtimeToTimestamp :: String -> Maybe Int
localtimeToTimestamp = fmap ( floor . utcTimeToPOSIXSeconds ) . parseTime defaultTimeLocale ( iso8601DateFormat $ Just "%H:%M:%S" )


withLazy :: Monad m => m a -> StateT ( Either ( m a ) a ) m b -> m b
withLazy = flip evalStateT . Left


getLazy :: Monad m => StateT ( Either ( m a ) a ) m a
getLazy = do
  s <- get
  case s of
    Left f -> do
      v <- lift f
      put $ Right v
      return v
    Right v ->
      return v
